package Alexandria::CDWS::Controller::Attachment;

use Moose;
use namespace::autoclean;

use XML::LibXML;
use Time::HiRes qw(gettimeofday tv_interval);
use Data::Dumper;
use File::Basename;
use IO::File::WithPath;

BEGIN {extends 'Catalyst::Controller::REST'; }

=head1 NAME

Alexandria::CDWS::Controller::Attachment - Catalyst Controller for attachments

=head1 DESCRIPTION

Catalyst Controller handling requests at
  /attachment

=head1 METHODS

=cut

sub _file_exists {
  my ($self, $c) = @_;

  if( $c->model('Attachments')->{roots} ) {
    # we check multiple destinations
    my $roots = $c->model('Attachments')->{roots};
    foreach my $ord ( sort keys %{ $roots } ) {
      my $dest = sprintf("%s/%s",
                     $roots->{ $ord },
                     $c->request->parameters->{path});
      if( -s "$dest" ) {

        my ( $name, $path, $suffix ) = fileparse( "$dest", 'tif', 'TIF' );

        #### ALPA-2758 corrupt JP tiffs < 20000101
        if( $c->request->parameters->{path} =~ /^\/JP/i && $suffix =~ /tif/i ) {
          open( F, "< $dest" );
          binmode F;
          local $/ = undef;
          my $data = <F>;
          close F;

          if( $data =~ /^II2/ ) {
            # bad
            $dest = "/tmp/$name$suffix";
            open( G, "> $dest" );
            binmode G;
            $data =~ s/^II2/II*/;
            print G $data;
            close G;
          }
        }
        return $dest;
      }
    }
    return undef; # not found
  }


  # NOTE: c->model('Attachments')->{root}
  # is the directory attachments physically live
  # <root>/EP/...
  my $file = sprintf("%s/%s",
                   $c->model('Attachments')->{root},
                   $c->request->parameters->{path});
  if(-s $file) {
    return $file;
  }
  return undef;
}

sub list : Local : ActionClass('REST') { }
sub list_GET {
  my ( $self, $c ) = @_;

  if( !$c->request->parameters->{ucid} ) {
    return $self->status_bad_request( $c, message => "missing required 'ucid' parameter" );
  }

  # timing information
  my $t = [gettimeofday];

  my $params = $c->request->parameters;

  my $ucid = $params->{ucid};
  my $attachments = $c->model('Attachments')->list($ucid);
  return $self->status_ok($c,
    entity => { status => 'success',
                time => sprintf( "%.6f", tv_interval($t) ),
                count => scalar( @{$attachments} ),
                attachments => $attachments
              }
  ) if scalar( @{$attachments} );

  return $self->status_no_content($c);
}

sub fetch : Local : ActionClass('REST') { }
sub fetch_GET {
  my ( $self, $c ) = @_;

  if( !$c->request->parameters->{path} ) {
    return $self->status_bad_request( $c, message => "missing required 'path' parameter" );
  }

  # timing information
  my $t = [gettimeofday];

  my $params = $c->request->parameters;
  my $user = $c->user;

  if( $c->request->parameters->{path} =~ /^\/CN/i
      && $user->subscription_level ne 'Premium-Plus' ) {
    return $self->status_forbidden($c, message => 'Access Denied: Please contact a sales representative (+1 203 779 5301).');
  }

  my $file = $self->_file_exists( $c );
  if( !$file ) {
    return $self->status_not_found($c, message => "$c->request->parameters->{path} not found");
  }
  my $types = MIME::Types->new();
  my $mime = $types->mimeTypeOf( $file );
  my $type;
  if( !$mime ) { $type = 'application/octet-stream'; }
  else         { $type = $mime->type(); }
  my $fh = IO::File::WithPath->new( $file, 'r' );
  $c->response->body($fh);
  $c->response->headers->content_type( $type );
  $c->response->headers->content_length( -s $file );
  $c->response->headers->last_modified( (stat($file))[9] );

  if( $file =~ /^\/tmp/ ) {
    $c->log->info( "DELETING $file" );
  }
}

sub serialize_list {
  my ($self, $c) = @_;

  my $response = XML::LibXML::Document->new('1.0');
  my $root = XML::LibXML::Element->new('response');
  $response->setDocumentElement($root);

  my $stash_key = (
          $self->{'serialize'} ?
                $self->{'serialize'}->{'stash_key'} :
                $self->{'stash_key'}
        ) || 'rest';
    my $ref = $c->stash->{$stash_key};
    # list ...
    if($ref && scalar(@{$ref}) > 0) {
      my $attachments = $root->addNewChild('', 'attachments');
      $attachments->setAttribute('count', scalar(@{$ref}));
      foreach my $a (@{$ref}) {
        my $attachment = $attachments->addNewChild('', 'attachment');
        $attachment->setAttribute('filename', $a->{filename});
        $attachment->setAttribute('pkey', $a->{pkey});
        $attachment->setAttribute('media', $a->{media});
        $attachment->setAttribute('path', $a->{path});
        $attachment->setAttribute('size', $a->{size});
      }
    }
    $c->response->output($response->toString(1));
}

sub serialize : ActionClass('Serialize') {}
sub end :Private {
  my($self, $c) = @_;
  $self->{request_end} = [gettimeofday];

  if( $c->request->path =~ /fetch/) {
    # no op
    return;
  }
  $c->forward('serialize');
}
=head1 AUTHOR

David Santamauro E<lt>david.santamauro@fairviewresearch.comE<gt>

=head1 LICENSE

This library is free software. You can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

__PACKAGE__->config(
  'default'   => 'application/json',
  'stash_key' => 'rest',
  'map'       => {

         'text/xml'           => 'XML::Alexandria::XResponse::XAttachment',
         'application/json'   => 'JSON::Alexandria',
         'text/x-json'        => 'JSON::Alexandria',

         'text/html'          => 'YAML::HTML',
         'text/x-data-dumper' => [ 'Data::Serializer', 'Data::Dumper' ],
         'text/x-data-denter' => [ 'Data::Serializer', 'Data::Denter' ],
         'text/x-data-taxi'   => [ 'Data::Serializer', 'Data::Taxi'   ],
         'application/x-storable'   => [ 'Data::Serializer', 'Storable' ],
         'application/x-freezethaw' => [ 'Data::Serializer', 'FreezeThaw' ],
         'text/x-config-general'    => [ 'Data::Serializer', 'Config::General' ],
         'text/x-php-serialization' => [ 'Data::Serializer', 'PHP::Serialization' ],
        }
);

__PACKAGE__->meta->make_immutable;

1;
