package Alexandria::CDWS::Controller::Citations;
use Moose;
use namespace::autoclean;

use JSON ();
use XML::LibXML;
use Data::Dumper;
use URI::Escape;
use Time::HiRes qw(gettimeofday tv_interval);

BEGIN {extends 'Catalyst::Controller::REST'; }

=head1 NAME

Alexandria::Controller::Citations - Main citations controller


=head1 DESCRIPTION

Search Catalyst Controller.

=head1 METHODS

=cut

use Data::Dumper;

sub _post_data {
  my ( $self, $c ) = @_;
  my $data;
  if(ref($c->req->body) =~ /File/) {
    my $tmp = $c->req->body;
    $tmp->seek(0,0);
    $data = do { local $/; <$tmp> };
  }
  else { $data = $c->req->body; }
  return $data;
}

sub _forward_citations {
  my ( $self, $c, $data ) = @_;
  $data =~ s/\n/ /g;  
  $data =~ s/,/ /g;   
  $data =~ s/\s+/ /g; 
  $data =~ s/^\s+//;
  $data =~ s/\s+$//;

  my $t = [gettimeofday];

  my @ucids = split(/\s+/, $data);
  my $user = $c->user;
  my $params = $c->request->parameters;

  my ($content_type) = grep { $self->{map}->{$_} } @{$c->request->accepted_content_types};

  my @ret;
  my %solropts = (
    qt => $c->model( $user->idx_model )->{params}->{qt},
    fl => 'ucid',
    rows => '1000',
    wt => 'json'
  );

  my $json = JSON->new->utf8->convert_blessed;
  my @responses = ();
  if(scalar(@ucids) > 1) {
    foreach my $ucid ( @ucids ) {
      my $obj = $json->decode(
        $c->model( $user->idx_model )->search(
          sprintf( "pcit:%s", $ucid ),
          \%solropts
        )->raw_response->content
      );
      # save root
      $obj->{root} = $ucid;
      push(@responses, $obj);
    }
  }
  else {
    my $obj = $json->decode(
      $c->model( $user->idx_model )->search(
        sprintf( "pcit:%s", $params->{ucid} ),
        \%solropts
      )->raw_response->content
    );
    # save root
    $obj->{root} = $params->{ucid};
    push(@responses, $obj);
  }

#  $c->log->debug( Dumper(\@responses) );

  my @fcites = ();
  foreach my $x ( @responses ) {
    my %s = ( root => $x->{root} );
    my @a = ();
    foreach my $doc ( @{$x->{response}->{docs} } ) {
      push(@a, $doc->{ucid});
    }

    if( $params->{extended} ) {
      $s{ucids} = $c->model( $user->rdb_model  )->_extended_data( \@a );
    }
    else {
      $s{ucids} = \@a;
    }
    push(@fcites, \%s );
  }
#$c->log->debug( Dumper( \@fcites ) );

  my %res = (
    status => 'success',
    time => sprintf('%.3f', tv_interval($t)),
    citations => \@fcites
  );
  return $self->status_ok( $c, entity => \%res );
}

sub forward : Local : ActionClass('REST') { }
sub forward_GET  :Args(0){
  my ( $self, $c ) = @_;

  my $params = $c->request->parameters;
  return $self->_forward_citations( $c, $params->{ucid} );
}

sub _backward_citations {
  my ( $self, $c, $data ) = @_;

  my $t = [gettimeofday];

  $data =~ s/\n/ /g;  
  $data =~ s/,/ /g;   
  $data =~ s/\s+/ /g; 
  $data =~ s/^\s+//;
  $data =~ s/\s+$//;

  my @ucids = split(/\s+/, $data);
  my $user = $c->user;
  my $params = $c->request->parameters;

  my ($content_type) = grep { $self->{map}->{$_} } @{$c->request->accepted_content_types};

  my @ret;

  # grab cites from postgres
  if(scalar(@ucids) > 1) {
    # array returned if multiple requested
    my @a = ();
    foreach my $ucid ( @ucids ) {
      my $cites = $c->model( $user->rdb_model  )->citations_backward( $ucid, $params->{extended} );
      push( @ret, $cites );
    }
  }
  else {
    # single structure
    push( @ret,
      $c->model( $user->rdb_model  )->citations_backward( $params->{ucid}, $params->{extended} )
    );
  }

  my %res = (
    status => 'success',
    time => sprintf('%.3f', tv_interval($t)),
    citations => \@ret
  );
  return $self->status_ok( $c, entity => \%res);
}

sub backward : Local : ActionClass('REST') { }
sub backward_GET  :Args(0){
  my ( $self, $c ) = @_;

  my $params = $c->request->parameters;
  if(!$params->{ucid}) {
    return $self->status_bad_request($c, message => 'insufficient parameters');
  }
  return $self->_backward_citations( $c, $params->{ucid} );
}

=head2 end

=cut

sub serialize : ActionClass('Serialize') {}
sub end :Private {
  my ($self, $c) = @_;
  # CAUTION: unknown side affects. Somehow, a POST
  # request, regardless of content-type and response->code,
  # sets the body to "Content-Type xxx is not supported"
  # For successul request/response, undef this because the
  # serializer sets the body
  if( $c->response->code == 200 && $c->response->has_body ) {
    $c->response->body(undef);
  }
  $c->forward('serialize');
}


=head1 AUTHOR

David Santamauro

=head1 LICENSE

This library is free software. You can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

__PACKAGE__->config(
  'default'   => 'application/json',
  'stash_key' => 'rest',
  'map'       => {

         'text/xml'           => 'XML::Alexandria::XResponse::Citations',
         #'text/xml'           => 'XML::Simple',
         'application/json'   => 'JSON::Alexandria',
         'text/x-json'        => 'JSON::Alexandria',

         'text/html'          => 'YAML::HTML',
         'text/x-data-dumper' => [ 'Data::Serializer', 'Data::Dumper' ],
         'text/x-data-denter' => [ 'Data::Serializer', 'Data::Denter' ],
         'text/x-data-taxi'   => [ 'Data::Serializer', 'Data::Taxi'   ],
         'application/x-storable'   => [ 'Data::Serializer', 'Storable' ],
         'application/x-freezethaw' => [ 'Data::Serializer', 'FreezeThaw' ],
         'text/x-config-general'    => [ 'Data::Serializer', 'Config::General' ],
         'text/x-php-serialization' => [ 'Data::Serializer', 'PHP::Serialization' ],
        }
);


__PACKAGE__->meta->make_immutable;

1;
