package Alexandria::CDWS::Controller::Custom::FWS;
use Moose;
use namespace::autoclean;

use Time::HiRes qw(gettimeofday tv_interval);
use JSON;
use XML::LibXML;
use XML::XML2JSON;

use Data::Dumper;
use Alexandria::Tools::AppNumFormat;
use Alexandria::Tools::DocNumFormat;

BEGIN {extends 'Catalyst::Controller::REST'; }

=head1 NAME

Alexandria::CDWS::Controller::Custom::ACWS - custom text web service


=head1 DESCRIPTION

Alexandria::CDWS::Controller::Custom::ACWS takes either a document ucid
or an application ucid (or permutations thereof) and returns a fws_application
structure based on application filing number. This service is application-number centric.

=cut

##### parsing methods
sub _should_parse {
  my ($self, $c, $fld) = @_;
  my $includes = $c->stash->{includes};
  my $ret = ( exists( $includes->{$fld} ) && $includes->{$fld} == 1 ) ? 1 : 0;
  return $ret;
}

sub _fws_ls_correspondence {
  my ( $self, $c, $ucids ) = @_;
  my $user = $c->user;

  if( !$ucids || !scalar( @{ $ucids } ) ) { return undef; }

  foreach my $ucid ( @{ $ucids } ) {
    # we assume there is only one EP member
    # and probe the A-level document
    if( $ucid =~ /^EP.*?A(1|2)$/ ) {
      return $c->model( $user->rdb_model )->fws_ls_corresponding_documents( $ucid );
    }
  }
}

sub _fws_ls_national_grants_information {
  my ( $self, $c, $ucid ) = @_;
  my $user = $c->user;
  return $c->model( $user->rdb_model )->fws_ls_national_grants_information( $ucid );
}

sub _parse_patent_document {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;

  my $fmt = $self->{_instruct}->{fmt};

$c->log->debug( '_parse_patent_document()' );

  foreach my $patdoc ( $root->getElementsByTagName('patent-document') ) {
    my $fam = $patdoc->getAttribute('family-id');
    if( $fam && $fam != -1 && !$output->{fam} ) { $output->{fam} = $fam; }

    # populate:
    #   - publicatons
    #   - grant
    #   - national-grant
    my $flag = $c->model('Custom')->app_or_grant(
                     $patdoc->getAttribute('country'),
                     $patdoc->getAttribute('kind'),
                     $patdoc->getAttribute('date'),
                     $patdoc->getAttribute('doc-number')
    );

    if( $flag eq 'G'
            && ( $fmt->{origin} && $fmt->{origin} eq 'EP' )
            && $patdoc->getAttribute('country') eq $fmt->{country} ) {
      $output->{national_grant} = $self->_fmt_document_id( { pn => $patdoc->getAttribute('ucid') }, 'p' );
      $output->{national_grant}->{date} = $patdoc->getAttribute('date');
      $output->{national_grant}->{lang} = $patdoc->getAttribute('lang')
        unless $patdoc->getAttribute('lang') eq 'XX';
    }
    elsif( $flag eq 'G' ) {
      $output->{grant} = $self->_fmt_document_id( { pn => $patdoc->getAttribute('ucid') }, 'p' );
      $output->{grant}->{date} = $patdoc->getAttribute('date');
      $output->{grant}->{lang} = $patdoc->getAttribute('lang')
        unless $patdoc->getAttribute('lang') eq 'XX';
    }
    else {
      my $publication = $self->_fmt_document_id( { pn => $patdoc->getAttribute('ucid') }, 'p' );
      $publication->{date} = $patdoc->getAttribute('date');
      $publication->{lang} = $patdoc->getAttribute('lang')
        unless $patdoc->getAttribute('lang') eq 'XX';
      push( @{ $output->{publications} }, $publication );
    }
  }
}

sub _parse_publication_reference {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;
  my $instruct = $self->{_instruct};

$c->log->debug( '_parse_publication_reference()' );
  # NOOP
}

sub _parse_application_reference_ex {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;
  my $instruct = $self->{_instruct};
  my @dates = ();
  my %apprefs = ();

$c->log->debug( '_parse_application_reference_ex()' );

  my ( $ucid, $docid, $epo_docid, $orig_docid ) = ();

  foreach my $appref ( $root->getElementsByTagName('application-reference') ) {

    # now ... there could be 2 @formats, epo|original
    foreach my $node ( $appref->getElementsByTagName('document-id') ) {
      if( $node->getAttribute('format') eq 'epo' )      { $epo_docid = $node; }
      if( $node->getAttribute('format') eq 'original' ) { $orig_docid = $node; }
    }
    if( $epo_docid ) { $docid = $epo_docid; }
    else             { $docid = $orig_docid; }

    $ucid = $appref->getAttribute('ucid');
    if( !$ucid ) { next; }
    push ( @{$apprefs{ $ucid }},
            { ctry => $docid->findvalue('./country'),
              anum => $docid->findvalue('./doc-number'),
              kind => ( $docid->findvalue('./kind') ) ? $docid->findvalue('./kind') : '',
              date => ( $docid->findvalue('./date') ) ? $docid->findvalue('./date') : '',
              lang => ( $docid->findvalue('./lang') ) ? $docid->findvalue('./lang') : ''
            }
    );
  }

  # now we have applicable application references
  my $id;

$c->log->debug( Dumper( $instruct->{fmt} ) );
$c->log->debug( "APPLICATIONS: " . Dumper( \%apprefs ) );
  # take the best one
  if( $apprefs{ $self->{_base_docid}->{an} } ) {
      $id = shift( @{$apprefs{ $self->{_base_docid}->{an} }} );
  }
  elsif( scalar( keys %apprefs ) == 1 ) {
    $id = shift( @{$apprefs{ $ucid }} );
  }
  else {
    foreach my $anucid ( sort keys %apprefs ) {
      my ( $ctry, $anum, $kind ) = split( /\-/, $anucid );
      # either we have the correct ctry+anum
      if( $instruct->{fmt}->{country} eq $ctry
            && $instruct->{fmt}->{app_number} eq $anum ) {
        $id = shift( @{$apprefs{$anucid}} ); last;
      }
      # ... or we try to grab origin+anum
      elsif( $instruct->{fmt}->{origin}
            && $instruct->{fmt}->{origin} eq $ctry
            && $instruct->{fmt}->{app_number} eq $anum ) { 
        $id = shift( @{$apprefs{$anucid}} ); last;
      }
    }
  }
  $output->{ucid} = sprintf( "%s-%s-%s", $id->{ctry}, $id->{anum}, $id->{kind} );
  $output->{country} = $id->{ctry};
  $output->{doc_number} = $id->{anum};
  $output->{kind} = $id->{kind};
  $output->{date} = $id->{date};
  $output->{lang} = $id->{lang} unless $id->{lang} eq 'XX';

#$c->log->debug( Dumper( \%apprefs ) );
#$c->log->debug( Dumper( $instruct ) );
$c->log->debug( "APPLICATION: " . Dumper( $id ) );
}

sub _parse_application_reference {
  my $self = shift;

  $self->_parse_application_reference_ex( @_ );


  my $c = shift;
  my $root = shift || return;
  my $output = shift;
  my $instruct = $self->{_instruct};
  my @dates = ();

$c->log->debug( '_parse_application_reference()' );

  foreach my $appref ( $root->getElementsByTagName('application-reference') ) {
    foreach my $docid ( $appref->getElementsByTagName('document-id') ) {
      my $ucid = $appref->getAttribute('ucid');
      # TODO: what if there isn't a DOCDB ucid?
      if( !$ucid || $ucid =~ /D$/ ) { next; }
      my ( $ctry, $anum, $kind ) = split( '-', $ucid );

      # national filing of input id
      # so, this will fail if the national data
      # for EP origin is missing
      # E.g., id=FR03251195&origin=EP
      if( $instruct->{fmt}->{country} eq $ctry ) {
        $output->{ucid} = $ucid if !$output->{ucid};
        $output->{country} = $docid->findvalue('./country')
          if $docid->findvalue('./country') && !$output->{country};
        $output->{doc_number} = $docid->findvalue('./doc-number')
          if $docid->findvalue('./doc-number') && !$output->{doc_number};
        $output->{kind} = $docid->findvalue('./kind')
          if $docid->findvalue('./kind') && !$output->{kind};
        $output->{date} = $docid->findvalue('./date')
          if $docid->findvalue('./date') && !$output->{date};
        $output->{lang} = $docid->findvalue('./lang')
          if $docid->findvalue('./lang') && !$output->{lang};
        $output->{format} = 'ifi';
      }

      if( $docid->findvalue('./date') ) {
        push( @dates, $docid->findvalue('./date') );
      }

    }
  }
  $output->{effective_filing_date} = ( sort @dates )[0];
  if( $output->{effective_filing_date} && !$output->{date} ) {
    $output->{date} = $output->{effective_filing_date};
  }
}

sub _parse_international_convention_data {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;
  my $instruct = $self->{_instruct};
$c->log->debug( '_parse_international_convention_data()' );

  foreach my $icd ( $root->getElementsByTagName('international-convention-data') ) {
    foreach my $pct ( $root->getElementsByTagName('pct-or-regional-filing-data') ) {
      if( !$output->{origin} ) { $output->{origin} = 'WO'; }
      my $ned = $pct->getElementsByTagName('national-entry-date')->[0];
      if( $ned  && !$output->{national_entry_date} ) {
        $output->{national_entry_date} = $ned->textContent();
      }
      my $pct_filed = $pct->findvalue('./document-id/date');
      if( $pct_filed && $pct_filed ne '' ) {
        # regardless of which is earlier, PCT filed trumps all other alternatives
        $output->{effective_filing_date} = $pct_filed;
        $self->{_effective_is_pct} = 1;
      }
    }
  }
}

sub _parse_priority {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;
  my $idx = 0;
  my @pr = ();
$c->log->debug( '_parse_priority()' );

  # find earliest
  my %priorities = ();
  foreach my $priority  ( $root->getElementsByTagName('priority-claim') ) {
    if( !$priority->getAttribute('ucid') ) { next; }
    my %docid = (
      ucid => $priority->getAttribute('ucid'),
      date => $priority->findvalue('./document-id/date')
    );
    $priorities{ $docid{date} . ':' . $docid{ucid} } = \%docid;
  }
#$c->log->debug( Dumper( \%priorities ) );

 # shift off earliest ...
 my $p = (sort keys %priorities)[0];
#$c->log->debug( Dumper( $priorities{$p} ) );
    
  # take first
  my $pris = $root->getElementsByTagName('priority-claims')->[0];
  if( $pris ) {
    foreach my $pri ( $pris->getElementsByTagName('priority-claim') ) {
      next if $pri->getAttribute('load-source') ne 'docdb';
      $idx++;
      my %s = (
          ucid => $pri->getAttribute('ucid'),
          country => $pri->findvalue('./document-id/country'),
          doc_number => $pri->findvalue('./document-id/doc-number'),
          kind => $pri->findvalue('./document-id/kind'),
          date => $pri->findvalue('./document-id/date')
      );
      push( @pr, \%s );

      if( $s{kind} eq 'W' && !$output->{origin} ) {
        $output->{origin} = 'WO';
      }
    }
  }
  if($idx == 0) { return; }

  my @sorted = sort {
    $a->{date} <=> $b->{date}
  } @pr;

  while( my $earliest = shift( @sorted ) ) {
    # assign first one
    $output->{priority} = $earliest
      if !$output->{priority};
  }
}

sub _parse_term_extension {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;
$c->log->debug( '_parse_term_extension()' );

  # take first ( any, really )
  my $pta = $root->getElementsByTagName('us-term-extension')->[0];
  if( $pta ) { $output->{pta} = $pta->textContent(); }
}

sub _parse_dates_of_public_availability {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;
  foreach my $patdoc ( $root->getElementsByTagName('patent-document') ) {
    if( my $dpa = $patdoc->getElementsByTagName('dates-of-public-availability')->[0] ) {
      if( !$output->{effective_grant_date} ) {
        if( my $rd = $dpa->getElementsByTagName('registered-date')->[0] ) {
          $output->{effective_grant_date} = $rd->textContent();
$c->log->debug( "EFFECTIVE GRANT DATE (dpa): $output->{effective_grant_date}" );
        }
      }
    }
  }
}

sub _parse_effective_grant {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;
  my $user = $c->user;

$c->log->debug( '_parse_effective_grant()' );

  $self->_parse_dates_of_public_availability( $c, $root, $output );
  if( !$output->{effective_grant_date} ) {
    foreach my $patdoc ( $root->getElementsByTagName('patent-document') ) {

      if( $patdoc->getAttribute('country') eq 'JP' ) {
        $output->{effective_grant_date} =
            $c->model( $user->rdb_model )->fws_ls_effective_grant_date( $patdoc->getAttribute('ucid') );
      }
      if( !$output->{effective_grant_date} && ( $output->{origin} && $output->{origin} eq 'WO' ) ) {
        if( $patdoc->getAttribute('country') eq 'WO' ) {
          my $grants = $self->_fws_ls_national_grants_information( $c, $patdoc->getAttribute('ucid') );
          foreach my $grant ( @{ $grants } ) {
            if( $output->{country} eq $grant->{ctry} && $grant->{date} ) {
              $output->{effective_grant_date} = $grant->{date};
$c->log->debug( "EFFECTIVE GRANT DATE (ls): $output->{effective_grant_date}" );
            }
          }
        }
      }
    }
  }
}

sub _parse_titles {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;

$c->log->debug( '_parse_titles()' );

  my %ttl = ();
  foreach my $patdoc ( $root->getElementsByTagName('patent-document') ) {
    foreach my $ttl ( $patdoc->getElementsByTagName('invention-title') ) {
      push( @{ $ttl{ $ttl->getAttribute('lang') } },
          { 'ref-ucid' => $patdoc->getAttribute('ucid'), text => $ttl->textContent() }
      );
    }
  }

#$c->log->debug( Dumper( \%ttl ) );

  # now, pop off first of each language
  foreach my $lang ( sort keys %ttl ) {
    my $ts = shift( @{$ttl{ $lang }} );
    push( @{ $output->{titles} },
      { 'ref-ucid' => $ts->{'ref-ucid'},
        'lang' => $lang,
        'title' => $ts->{text} }
    );
  }
}

sub _parse_related_documents {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;

$c->log->debug( '_parse_related_documents()' );

  my @dates = (); # for effective date calculation
  foreach my $reldocs ( $root->getElementsByTagName('related-documents') ) {
    foreach my $reldoc ( $reldocs->getElementsByTagName('relation') ) {
      if( $reldoc->getAttribute('type') =~ /(continuation|division)/) {
#$c->log->debug("Relation ( " . $reldoc->getAttribute('type') . " ): "
#                  . $reldoc->findvalue('./parent-doc/@ucid')
#                  . "( "
#                  . $reldoc->findvalue('./parent-doc/document-id/date')
#                  . " )"
#);
        if( !$output->{relation} ) {
          $output->{relation} = $reldoc->getAttribute('type');
          $output->{parent} = $reldoc->findvalue('./parent-doc/@ucid');
        }
        if( $reldoc->findvalue('./parent-doc/document-id/date') ) {
          push( @dates, $reldoc->findvalue('./parent-doc/document-id/date') );
        }
      }
    }
    if( $output->{effective} ) {
      push( @dates, $output->{effective} );
    }
    push( @dates, $output->{date} );
    $output->{effective} = ( sort @dates )[0]
      if !$self->{_effective_is_pct} ;
  }
}

sub _parse_ifi_integrated_content {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;

$c->log->debug( '_parse_ifi_integrated_content()' );

  # take first
  my $ifi = $root->getElementsByTagName('ifi-integrated-content')->[0];
  if( $ifi ) {
    my $psta = $ifi->getElementsByTagName('ifi-patent-status-description')->[0];
    my $claims = $ifi->getElementsByTagName('ifi-claims')->[0];
    my $pte = $ifi->getElementsByTagName('ifi-term-extension')->[0];
    my $stdname = $ifi->getElementsByTagName('ifi-standardized-name-current')->[0];

    my @claims_hierarchy = ();
    if( $claims ) {
      foreach my $clm ( $claims->getElementsByTagName('ifi-claim') ) {
        my @claim_refs = ();
        foreach my $ref ( $clm->getElementsByTagName('ifi-claim-ref') ) {
          push( @claim_refs, $ref->getAttribute('num') );
        }
        push(
          @claims_hierarchy,
          {
            num => $clm->getAttribute('num'),
            type => $clm->getAttribute('type'),
            refs => \@claim_refs
          }
        );
      }
    }
    my %ifidata = (
      claims_total => ( $claims ) ? $claims->getAttribute('total') : undef,
      claims_independent => ( $claims ) ? $claims->getAttribute('independent') : undef,
      claims_hierarchy => ( scalar( @claims_hierarchy ) ) ? \@claims_hierarchy : undef,
      pte => ( $pte ) ? $pte->textContent() : undef,
      assignee => ( $stdname ) ? $stdname->findvalue('./addressbook/name') : undef,
    );
    if( $psta ) {
      my $ps = $psta->textContent();
      if( $ps =~ /^(Pend|Grant|Active)/ ) {
        $ifidata{patent_status} = 'ALIVE';
      }
      else {
        $ifidata{patent_status} = 'LAPSED';
      }
    }
#$c->log->debug( Dumper( $ifidata{claims_hierarchy} ) );
    foreach my $key ( sort keys %ifidata ) {
      if ($ifidata{$key} ) { $output->{$key} = $ifidata{$key}; }
    }
  }
}

sub _parse_parties {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;

$c->log->debug( '_parse_parties()'); 

  my $parties = $root->getElementsByTagName('parties')->[0] || return;

  if( !$output->{assignee} ) {
    # take first
    # if we don't have assignee already, pick:
    #   docdb, if not docdb then original
    my %as;
    foreach my $asg ( $parties->getElementsByTagName('assignee') ) {
      $as{ $asg->getAttribute('format') } =
        ( $asg->findvalue('./addressbook/name') )
            ? $asg->findvalue('./addressbook/name')
            : $asg->findvalue('./addressbook/last-name');
      last; # take first one
    }
#$c->log->debug( Dumper( \%as ) );
    if( $as{epo} )       { $output->{assignee} = $as{epo}; }
    elsif( $as{original} ) { $output->{assignee} = $as{original}; }
  
    if( !$output->{assignee} ) {
#$c->log->debug( $parties->toString(1) );
      # grab first docdb applicant
      foreach my $asg ( $parties->getElementsByTagName('applicant') ) {
        if( $asg->getAttribute('format') eq 'epo' ) {
          $output->{assignee} =
            ( $asg->findvalue('./addressbook/name') )
                ? $asg->findvalue('./addressbook/name')
                : $asg->findvalue('./addressbook/last-name');
          last;
        }
      }
    }
  }

  if( $self->_should_parse( $c, 'inventors') ) {
    my %in;
    foreach my $inv ( $parties->getElementsByTagName('inventor') ) {
#$c->log->debug( $inv->toString(1) );
      my $n = ( $inv->findvalue('./addressbook/name') )
              ? $inv->findvalue('./addressbook/name')
              : $inv->findvalue('./addressbook/last-name') . ', ' . $inv->findvalue('./addressbook/first-name');
      $n =~ s/,//g;
      $n =~ s/;//g;
      $n =~ s/\.//g;
      $n =~ s/^\s+//;
      $n =~ s/\s+$//;
      $in{ uc($n) }++;
    }
#$c->log->debug( Dumper( \%in ) );
    foreach my $n ( sort keys %in ) {
      push( @{ $output->{inventors} }, $n );
    }
  }
}

sub _parse_classifications {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;

$c->log->debug( '_parse_classifications()' );

  # take first
  my $techical_data = $root->getElementsByTagName('technical-data')->[0];
  # grab all /sigh/
  # ic, uc, cpc, fi, fterms, ecla
  if( !$techical_data ) { return; }

  my %ipc;
  my %ipcr;
  my %cpc;
  my %nc;
  my %fi;
  my %ft;
  my %ecla;

  my $elements_a = {
    'classification-ipcr' => { type => 'ipcr', struct => \%ipcr },
    'classification-cpc' => { type => 'cpc', struct => \%cpc },
    'classification-ecla' => { type => 'ecla', struct => \%ecla },
    'f-term' => { type => 'ft', struct => \%ft },
  };
  # these have <main|further>
  my $elements_b = {
    'classification-ipc' => { type => 'ipc', struct => \%ipc },
    'classification-national' => { type => 'nc', struct => \%nc },
  };

  foreach my $key ( sort keys %{$elements_a} ) {
    my $name = $key;
    my $type = $elements_a->{$key}->{type};
    my $s = $elements_a->{$key}->{struct};
    
    foreach my $cls ( $techical_data->getElementsByTagName( $name ) ) {
      my $data = substr( $cls->textContent(), 0, 16 );
      $data =~ s/\s+$//;
      $s->{$data} = 1;
    }
    my @c = sort keys ( %{$s} );
    if( scalar( @c ) ) {
      push( @{ $output->{classifications} }, { $type => \@c } );
    }
  }

  foreach my $key ( sort keys %{$elements_b} ) {
    my $name = $key;
    my $type = $elements_b->{$key}->{type};
    my $s = $elements_b->{$key}->{struct};
    
    foreach my $contcls ( $techical_data->getElementsByTagName( $name ) ) {
      foreach my $sname ('main-classification', 'further-classification') {
        foreach my $cls ( $contcls->getElementsByTagName( $sname ) ) {
          my $data = substr( $cls->textContent(), 0, 16 );
          $data =~ s/\s+$//;
          $s->{$data} = 1;
        }
      }
    }
    my @c = sort keys ( %{$s} );
    if( scalar( @c ) ) {
      push( @{ $output->{classifications} }, { $type => \@c } );
    }
  }
#$c->log->debug( Dumper( $output->{classifications} ) );
}

sub _xml2json {
  my ($self, $xml) = @_;
  my $xjson = XML::XML2JSON->new();
  return $xjson->convert( $xml );
}

sub _parse_abstract {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;

$c->log->debug( '_parse_abstract()' );

  # original lang (assuming patent-document/@lang)
  my $lang = 'XX';
  my $pd = $root->getElementsByTagName('patent-document')->[0];
  if( $pd->getAttribute('lang') ) { $lang = $pd->getAttribute('lang'); }

  # english
  foreach my $ab ( $root->getElementsByTagName('abstract') ) {
    if( $ab->getAttribute('lang') =~ /en/i ) {
      $output->{abstract_en} = $ab->toString();
    }
    if( $ab->getAttribute('lang') =~ /$lang/i ) {
      $output->{abstract_original} = $ab->toString();
    }
  }
}

sub _parse_description {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;

$c->log->debug( '_parse_description()' );

  my $lang = 'XX';
  my $pd = $root->getElementsByTagName('patent-document')->[0];
  if( $pd->getAttribute('lang') ) { $lang = $pd->getAttribute('lang'); }

  foreach my $desc ( $root->getElementsByTagName('description') ) {
    if( $desc->getAttribute('lang') =~ /en/i ) {
      $output->{description_en} = $desc->toString();
    }
    if( $desc->getAttribute('lang') =~ /$lang/i ) {
      $output->{description_original} = $desc->toString();
    }
  }
}

sub _parse_claims {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;

$c->log->debug( '_parse_claims()' );

  # original lang (assuming patent-document/@lang)
  my $lang = 'XX';
  my $pd = $root->getElementsByTagName('patent-document')->[0];
  if( $pd->getAttribute('lang') ) { $lang = $pd->getAttribute('lang'); }

  # don't overwrite IFI data (if available)
  if( !$output->{claims_total} ) {
    # take first language
    my $claims = $root->getElementsByTagName('claims')->[0];
    if( !$claims ) { return; }
    foreach my $claim ( $claims->getElementsByTagName('claim') ) {
      $output->{claims_total}++;
    }
  }

  # english
  foreach my $claims ( $root->getElementsByTagName('claims') ) {
    if( $claims->getAttribute('lang') =~ /en/i ) {
#$c->log->debug( "Capturing claims_en ..." );
      $output->{claims_en} = $claims->toString();
    }
    if( $claims->getAttribute('lang') =~ /$lang/i ) {
#$c->log->debug( "Capturing claims_original ..." );
      $output->{claims_original} = $claims->toString();
    }
  }
#$c->log->debug( Dumper( $output->{claims_en} ) );
}

sub _parse_national_entry_date {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;
  my $fmt = $c->stash->{_fmt};

$c->log->debug( '_parse_national_entry_date()' );

  if( $output->{national_entry_date} ) {
    return;
  }
#$c->log->debug( Dumper( $output ) );
  foreach my $patdoc ( $root->findnodes("//patent-document[\@country='WO']") ) {
    # looking for national phase entries
    foreach my $event ( $patdoc->findnodes("//legal-event[\@code='WWE']") ) {
      my $worthy = 0;
      foreach my $value ( $event->getElementsByTagName('event-attribute-value') ) {
        # looking for country and number
        my $data = $value->textContent();
        if( !$data ) { next; }
        if( $output->{country} =~ /^$data/i ) { $worthy++; }
        if( $output->{doc_number} =~ /^$data/i ) { $worthy++; }
        if( $fmt && $fmt->{input} ) {
          my $x = $fmt->{input};
          $x =~ s/^$output->{country}//;
          $x =~ s/^\W//g;
          if ($x =~ /^$data/ ) { $worthy++; }
        }
      }
      if( $worthy >= 2 ) {
$c->log->debug( "national_entry_date => $event->{date}" );
        $output->{national_entry_date} = $event->{date};
      }
    }
  }
}

#### request methods
sub _post_params {
  my ( $self, $c ) = @_;
  my %params;

  if($c->req->content_type eq 'application/x-www-form-urlencoded') {
    my $s = $c->request->parameters;
    foreach my $key ( keys %{$s} ) {
      my $val = $s->{$key};
      $val =~ s/^\s+//;
      $val =~ s/\s+$//;
      $params{$key} = $val;
    }
  }
  else {
    my $data;
    if(ref($c->request->body) =~ /File/) {
      my $tmp = $c->request->body;
      $tmp->seek(0,0);
      $data = do { local $/; <$tmp> };
    }
    elsif($c->request->body) {
      $data = $c->request->body;
    }
    else {
      my @a = ();
      my $h = $c->request->parameters;
      foreach my $key ( keys %{$h} ) {
        if(ref($h->{$key}) =~ /ARRAY/) {
          foreach my $f ( @{$h->{$key}} ) {
            push(@a, "$key=$f");
          }
        }
        else { push(@a, "$key=$h->{$key}"); }
      }
      $data = join('&', @a);
    }
    my $str = uri_unescape($data);
    $str =~ s/^\s+//;
    $str =~ s/\s+$//;
    my @a = split(/\&/, $str);
    foreach my $e (@a) {
      my ($key, $val) = split(/=/, $e);
      $params{$key} = $val;
    }
  }
  return \%params;
}

sub _build_response {
  my( $self, $c, $fmt, $patent_documents ) = @_;
  my %o = ();
  my $p = XML::LibXML->new( huge => 1 );

  my $dom;
  eval { $dom = $p->parse_string( $patent_documents ); };
  if( $@ ) {
    $c->log->error( $@ );
    return undef;
  }

  my $params = $c->request->parameters;
  if( $params->{origin} ) { $o{origin} = $params->{origin}; }

  my $root = $dom->documentElement();
  $self->_parse_patent_document( $c, $root, \%o );
  $self->_parse_publication_reference( $c, $root, \%o );
  $self->_parse_application_reference( $c, $root, \%o );
  $self->_parse_international_convention_data( $c, $root, \%o );
  $self->_parse_national_entry_date( $c, $root, \%o );
#  $self->_parse_related_documents( $c, $root, \%o );
##  $self->_parse_priority( $c, $root, \%o );
#  $self->_parse_term_extension( $c, $root, \%o );
##  $self->_parse_dates_of_public_availability( $c, $root, \%o );
##  $self->_parse_effective_grant( $c, $root, \%o );
#  if( $o{grant} && $o{grant}->{date} && !$o{effective_grant_date} ) {
#    $o{effective_grant_date} = $o{grant}->{date};
#  }
#  $self->_parse_titles( $c, $root, \%o );
#  $self->_parse_ifi_integrated_content( $c, $root, \%o );
#  $self->_parse_parties( $c, $root, \%o );
#  if( $self->_should_parse( $c, 'classifications') ) {
#    $self->_parse_classifications( $c, $root, \%o );
#  }
#  if( $self->_should_parse( $c, 'abstract_en')
#        || $self->_should_parse( $c, 'abstract_original') ) {
#    $self->_parse_abstract( $c, $root, \%o );
#  }
#
#  if( $self->_should_parse( $c, 'description_en')
#        || $self->_should_parse( $c, 'description_original') ) {
#    $self->_parse_description( $c, $root, \%o );
#  }
#
#  # slightly more complicated than it should be ...
#  if( $self->_should_parse( $c, 'claims_en' )
#        || $self->_should_parse( $c, 'claims_original')
#        || ( $o{country} ne 'US' && $self->_should_parse( $c, 'claims_total') ) ) {
#    $self->_parse_claims( $c, $root, \%o );
#  }
#  if( $json_obj->{simple_family} ) {
#    my @a = sort( @{ $json_obj->{simple_family} } );
#    $o{simple_family} = \@a;
#  }
#  if( $self->_should_parse( $c, 'extended_family' ) ) {
#    my $pd = $root->getElementsByTagName('patent-document')->[0];
#    my $ef = $c->model('MSSQL')->extended_inpadoc_family( $pd->getAttribute('ucid') );
#    my @a = sort( @{ $ef } );
#    $o{extended_family} = \@a;
#  }

  return \%o;
}

sub _search {
  my ( $self, $c, $q, $solrparams ) = @_;
  my $user = $c->user;

  my $model = $c->model( $user->idx_model );

  if( !$solrparams->{qt} ) { $solrparams->{qt} = $model->{params}->{qt}; }

$c->log->debug( "SEARCH: $q" );

  my $res = $model->search( $q, $solrparams );
  my $json = JSON->new->utf8->convert_blessed;
  my $obj;
  eval {
    $obj = $json->decode( $res->raw_response->content );
  };
  if( $@ ) {
    $c->log->error( 'JSON error: ' . $@ );
    return undef;
  }
  return $obj;
}

sub _build_search_us_app {
  my ( $self, $c, $fmt, $input_is_ucid ) = @_;

  if( $input_is_ucid ) {
    return "( an: $fmt->{ucid} )";
  }

  if( $c->request->parameters->{date} ) {
    # trust the ucid
    my $a = $fmt->ucid;
    $a =~ s/\-XX$//;
    return "an:$a";
  }
  else {
    # let the guessing games begin ...
    my @a = ();
    if( $fmt && $fmt->{series} && $fmt->{series} >= 13) {
      @a = ( sprintf( "anctry: %s", $fmt->country ),
             sprintf( "annum: ????%02d%06d", $fmt->series, $fmt->app_number)
      );
    }
    else {
      my $anum = sprintf("%d", $fmt->app_number);
      @a = ( sprintf( "anctry: %s", $fmt->country ),
             sprintf( "annum: %d", $fmt->app_number)
      );
      my $t = ( 6 - length( $anum ) + 1 );
      if( $t == 1 ) { $a[1] .= '??'; }
      else { for(my $i = 0; $i < $t && $i < 2; $i++) { $a[1] .= '?'; } }
    }
    if( $fmt->{series} ) {
      push( @a, sprintf( "anseries: %02d", $fmt->{series} ) );
    }
    return sprintf("%s AND ( ( adyear: [1700 TO *] ) )", join( ' AND ', @a ) );
  }
}

sub _build_search {
  my ( $self, $c, $fmt, $endpoint, $input_is_ucid ) = @_;
  # short-circuit US
  if( $fmt->{country} eq 'US' && $endpoint eq 'an' ) {
    return $self->_build_search_us_app( $c, $fmt, $input_is_ucid );
  }
  if( $fmt && $fmt->{alt} ) {
    return sprintf("( %s: (%s OR %s) ) AND ( ( %s: %s ) ) AND ( ( adyear: [1700 TO *] ) )",
               ( $endpoint eq 'an' ) ? 'anctry' : 'pnctry',
                   $fmt->{country}, $fmt->{alt}->{country},
               ( $endpoint eq 'an' ) ? 'annum'  : 'pnnum',
               ( $endpoint eq 'an' ) ?  $fmt->{app_number} : $fmt->{doc_number}
    );
  }
  else {
    my $ucid = $fmt->{ucid};
    $ucid =~ s/\-XX$//i;
    return sprintf("( ( %s: %s ) AND ( adyear: [1700 TO *] ) ) %s",
                $endpoint,
                $ucid,
                ( $endpoint eq 'an' ) ? 'NOT ( ankind:W )' : ''
    );
  }
}

sub _fmt_document_id {
  my ( $self, $struct, $fflag ) = @_;
  my %docid = ();
  my ($ctry, $dnum, $kind) = split( /\-/, $struct->{ $fflag . 'n' } );
  return {
    ucid => $struct->{ $fflag . 'n' },
    country => $ctry,
    doc_number => $dnum,
    kind => $kind,
    date => $struct->{ $fflag . 'd' }
  };
}

sub _fws_application_from_fam {
  my ( $self, $c, $fmt, $struct ) = @_;
  my $user = $c->user;

#$c->log->debug( Dumper( $struct ) );
#$c->log->debug( Dumper( $fmt ) );

  # we assume the best
  my %meta = (
    iid => $fmt->{input},
    lookup_status => 'found',
    confidence => 100,
  );

  # the keys to {struct} are family IDs
  my @keys = keys %{ $struct };

  # if there is more than one family, found-duplicate
  # NOTE: this is not quite right when e.g., a US pregrant
  # has a family id but its newly-published grant doesn't
  if( scalar( @keys ) > 1 ) {
    if( !defined( $struct->{-1} ) ) {
      $c->log->info( "multiple family-id(s) found, returning found-duplicate");
      $meta{lookup_status} = 'found-duplicate';
      return \%meta;
    }
  }

  # pluck first valid family-id
  my $fam;
  foreach my $f ( @keys ) { if( $f != -1 ) { $fam = $f; } }
  if( !$fam ) {
    $meta{lookup_status} = 'not-found';
    return \%meta;
  }

#$c->log->debug( Dumper( $struct->{$fam} ) );
  # grab applicable base docid
  my $base_docid;
  # are we dealing with /dn or /an
  if( $fmt->{app_number} ) {
    foreach my $docid ( @{ $struct->{$fam} } ) {
      if( $fmt->{origin} && $fmt->{origin} eq 'EP' ) {
        if( $docid->{an} =~ /^$fmt->{alt}->{country}/
              && $docid->{an} =~ /$fmt->{alt}->{app_number}/ ) {
          $base_docid = $docid; last;
        }
      }
      else {
        if( $docid->{an} =~ /^$fmt->{country}/
              && $docid->{an} =~ /$fmt->{app_number}/ ) {
          $base_docid = $docid;
          # best one ... bail
          last;
        }
        elsif( $docid->{an} =~ /^$fmt->{alt}->{country}/
               &&  $docid->{an} =~ /$fmt->{alt}->{app_number}/ ) {
          $base_docid = $docid;
        }
      }
    }
  }
  else {
    foreach my $docid ( @{ $struct->{$fam} } ) {
      if( $fmt->{origin} && $fmt->{origin} eq 'EP' ) {
        if( $docid->{pn} =~ /^$fmt->{alt}->{country}/
              && $docid->{pn} =~ /$fmt->{alt}->{doc_number}/ ) {
          $base_docid = $docid; last;
        }
      }
      else {
        if( $docid->{pn} =~ /^$fmt->{country}/
              && $docid->{pn} =~ /$fmt->{doc_number}/ ) {
          $base_docid = $docid; last;
        }
        elsif( $docid->{pn} =~ /^$fmt->{alt}->{country}/
               &&  $docid->{pn} =~ /$fmt->{alt}->{doc_number}/ ) {
          $base_docid = $docid;
        }
      }
    }
  }
$c->log->debug( sprintf("FETCH: %s, %s, %s",
           $fam, $fmt->{country}, ( $fmt->{origin} ) ? $fmt->{origin} : "NULL")
);
$c->log->debug( Dumper( $base_docid ) );

  $self->{_base_docid} = $base_docid;

  my $members = $c->model( $user->rdb_model )->fws_members(
          $fam, $fmt->{country}, $base_docid->{pn}, $base_docid->{an}, $fmt->{origin} );
  my @patent_documents = ('<patent-documents>');
  foreach my $member ( @{ $members } ) {
#$c->log->debug( "RECV: $member->{ucid}" );
    push( @patent_documents, $member->{content} );
  }
  push( @patent_documents, '</patent-documents>');
  $meta{ fws_container } = $self->_build_response( $c, $fmt, join('', @patent_documents, ) );
  return \%meta;
}

sub _fws_application_from_dn {
  my ( $self, $c, $fmt ) = @_;

  $c->stash( _fmt => $fmt );

  my $q = $self->_build_search( $c, $fmt, 'pn' );
  my %solrparams = (
    'fl' => 'ucid,an,fam',
    'rows' => 10,
  );

  my $obj = $self->_search( $c, $q, \%solrparams );
  my %family = ();

  if( $obj && $obj->{response} ) {
    foreach my $doc ( @{ $obj->{response}->{docs} } ) {
      push( @{ $family{ $doc->{fam} } },
          { ucid => $doc->{ucid},
            pnctry => substr( $doc->{ucid}, 0, 2),
            an => $doc->{an},
            pn => $doc->{ucid},
          }
      );
    }
  }
  return $self->_fws_application_from_fam( $c, $fmt, \%family );
}

sub _fws_application_from_an {
  my ( $self, $c, $fmt, $input_is_ucid ) = @_;

  $c->stash( _fmt => $fmt );

  my $q = $self->_build_search( $c, $fmt, 'an' );
  my %solrparams = (
    'fl' => 'ucid,an,fam',
    'rows' => 100,
  );

  my $obj = $self->_search( $c, $q, \%solrparams );
  my %family;
  my @ucids = ();

  if( $obj && $obj->{response} ) {
    foreach my $doc ( @{ $obj->{response}->{docs} } ) {
      push(@ucids, $doc->{ucid} );
      push( @{ $family{ $doc->{fam} } },
          { ucid => $doc->{ucid},
            pnctry => substr( $doc->{ucid}, 0, 2),
            an => $doc->{an},
            pn => $doc->{ucid},
          }
      );
    }
  }
  return $self->_fws_application_from_fam( $c, $fmt, \%family );
}


# valid parameters:
#   (ur) uniqueResults => [0|1]
#     Return one unique result (TODO: how to choose whic one?)
#   (ct) confidenceThreshold => <int>
#     TODO: define factors
#   (nf) notFound => [0|1]
#     Return list of input numbers that have no match
#   (dp) duplicates
#     Return only results where input has multiple matches
#   (ir) includeRequired => [0|1]
#     Inlcude required field (TODO: what is required field?)
#   (it) includeText => [0|1]
#     Include optional text fields (TODO: define optional vs required)
#   (if) includeField  => fl[,fl,fl]
#     Only out fields listed (TODO: define field names)
#   (ef) excludeField => fl[,fl,fl]
#     Suppress output of these fields
#   (dd) deduplicate
#     remove identical responses (TODO: define identical)
sub _handle_includes {
  my ( $self, $c ) = @_;
  my %required = (
    ucid => 1,
    country => 1,
    doc_number => 1,
    kind => 1,
    date => 1,
    lang => 1,
    format => 1,
    effective => 1,
    effective_filing_date => 1,
    origin => 1,
    publications => 1,
    grant => 1,
    national_entry_date => 1,
    effective_grant_date => 1,
    designated_states => 1,
    national_grant => 1,
    pta => 1,
    terminal_disclaimer => 1, # TODO
    relation => 1, parent => 1,
    examination_request_date => 1, # TODO
    priority => 1,
    patent_status => 1,
    assignee => 1,
    titles => 1,
    fam => 1,
    corresponding_document => 1,
  );
  my %optional = (
    inventors => 0,
    simple_family => 0,
    extended_family => 0,
    classifications => 0,
    claims_total => 0,
    claims_hierarchy => 0,
    claims_en => 0,
    claims_original => 0,
    abstract_en => 0,
    abstract_original => 0,
    description_en => 0,
    description_original => 0,
  );
  my %includes = (
    %required
  );

  my $params = $c->request->parameters;
  if( $params->{'include-only'} ) {
    # special
    my %s = ( ucid => 1 );
    if( ref( $params->{'include-only'} ) =~ /ARRAY/) {
      foreach my $opt ( @{ $params->{'include-only'} } ) {
        $s{$opt} = 1;
      }
    }
    else {
      my @a = split( /,/, $params->{'include-only'} );
      foreach my $io ( @a ) {
        $s{ $io } = 1;
      }
    }
#$c->log->debug( Dumper( \%s ) );
    return \%s;
  }

  if( $params->{'include-optional'} ) {
    foreach my $opt ( keys %optional ) { $includes{$opt} = 1; }
  }

  if( $params->{'include'} ) {
    if( ref( $params->{'include'} ) =~ /ARRAY/) {
      foreach my $opt ( @{ $params->{'include'} } ) {
        $includes{$opt} = 1;
      }
    }
    else {
      $includes{ $params->{'include'} } = 1;
    }
  }
  if( $params->{'exclude'} ) {
    if( ref( $params->{'exclude'} ) =~ /ARRAY/) {
      foreach my $opt ( @{ $params->{'exclude'} } ) {
        delete $includes{$opt};
      }
    }
    else {
      delete $includes{ $params->{'exclude'} };
    }
  }
  return \%includes;
}

sub _segment_raw_data {
  my ( $self, $c, $data ) = @_;
  my $str = $data;
  $str =~ s/\n/ /g;  # remove all new lines
  $str =~ s/\s+/ /g; # tighten up spaces
  $str =~ s/^\s+//;
  $str =~ s/\s+$//;
  # our only delimiter is the comma
  my @a = split( /,/, $str );
  return \@a;
}

sub _process_input_parameters {
  my ( $self, $c, $params, $endpoint ) = @_;

  my @x = ();
  my $ids = $self->_segment_raw_data( $c, $params->{id} );
  foreach my $id ( @{ $ids } ) {
    my %h = ( id => $id );
    if( $params->{origin} && $params->{origin} =~ /(EP|WO)/ ) {
      $h{origin} = $params->{origin};
    }
    if( $params->{date} )   { $h{date} = $params->{date}; }
    push( @x, \%h );
  }

  foreach my $param ( keys %{ $params } ) {
    if( $param =~ /^origin\.(\d)$/ ) {
      my $sub = $1 - 1;
      $x[ $sub ]->{origin} = $params->{ $param }
        unless $params->{ $param } !~ /(EP|WO)/;
    }
    elsif( $param =~ /^date\.(\d)$/ ) {
      my $sub = $1 - 1;
      $x[ $sub ]->{date} = $params->{ $param };
    }
  }

  foreach my $s ( @x ) {
    my %fmtparams = ( input => $s->{id} );
    if( $s->{date} )    { $fmtparams{date} = $s->{date}; }
    if( $s->{origin} )  { $fmtparams{origin} = $s->{origin}; }
    my $fmt;
    if( $endpoint eq 'an' )   { $s->{fmt} = Alexandria::Tools::AppNumFormat->new( \%fmtparams ); }
    elsif( $endpoint eq 'dn') { $s->{fmt} = Alexandria::Tools::DocNumFormat->new( \%fmtparams ); }
    else                      { $s->{fmt} = Alexandria::Tools::AppNumFormat->new( \%fmtparams ); }
  }
  return \@x;
}

=head1 METHODS

=cut

sub _process_request {
  my ( $self, $c, $params, $endpoint ) = @_;
  my $user = $c->user;

  my $t = [gettimeofday];
  my $data = $params->{id};
  $c->stash( are_ucids => $params->{is_ucid} );

  my ($content_type) = grep { $self->{map}->{$_} } @{$c->request->accepted_content_types};
#$c->log->debug( Dumper( $params ) );

  return $self->status_bad_request(
    $c, message => 'insufficient parameters'
  ) if !$data;

  my $includes = $self->_handle_includes( $c );
  $c->stash( includes => $includes ); # save for use everywhere
#$c->log->debug( Dumper( $includes ) );

  my $is = $self->_process_input_parameters( $c, $params, $endpoint );
#$c->log->debug( Dumper( $is ) );

  my @result = ();

  foreach my $in ( @{$is} ) {

    # for posterity ...
    $self->{_instruct} = $in;

    # both _fws_application_from_dn and _fws_application_from_an
    # return a \%meta hash structure
    if( $endpoint eq 'dn') {
      push( @result,
           $self->_fws_application_from_dn(
              $c, $in->{fmt}
           )
      );
    }
    elsif( $endpoint eq 'an' ) {
      push( @result,
           $self->_fws_application_from_an(
              $c, $in->{fmt}
           )
      );
    }
  }

  foreach my $container ( @result ) {
    my $fws_app = $container->{fws_container};
    foreach my $fld ( keys %{ $fws_app } ) {
      if( !exists( $includes->{$fld} ) || !$includes->{$fld} ) {
$c->log->debug( "deleting $fld ..." );
        delete $fws_app->{$fld};
      }
      else {
        if( ( !$content_type && $fld =~ /abstract|claims_en|claims_original|description_en|description_original/ )
                 || ( $content_type && $content_type !~ /xml/ && $fld =~ /abstract|claims_en|claims_original|description_en|description_original/ ) ) {
          eval { $fws_app->{$fld} = $self->_xml2json( $fws_app->{$fld} ); };
          if( $@ ) {
            $c->log->error( Dumper( $fws_app->{$fld} ) );
            $c->log->error( $@ );
          }
        }
      }
    }
  }
#$c->log->debug( Dumper( \%result ) );

  my %ret = (
    status => 'success',
    time => sprintf( "%.3f", tv_interval( $t ) ),
    method => "custom/fws/$endpoint",
    content => \@result,
  );
  return $self->status_ok( $c, entity => \%ret );
}

### document number end-point
sub dn : Local : ActionClass('REST') { }
sub dn_GET  :Args(0){
  my ( $self, $c ) = @_;
  return $self->_process_request ($c, $c->request->parameters, 'dn' );
}

sub dn_POST :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_process_request( $c, $self->_post_params( $c ), 'dn' );
}

### application number end-point
sub an : Local : ActionClass('REST') { }
sub an_GET  :Args(0){
  my ( $self, $c ) = @_;
  return $self->_process_request ($c, $c->request->parameters, 'an' );
}

sub an_POST :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_process_request( $c, $self->_post_params( $c ), 'an' );
}

### anaqua-specific application number end-point
sub aqs : Local : ActionClass('REST') { }
sub aqs_GET  :Args(0){
  my ( $self, $c ) = @_;
  return $self->_process_request ($c, $c->request->parameters, 'aqs' );
}

sub aqs_POST :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_process_request( $c, $self->_post_params( $c ), 'aqs' );
}


=head2 end

=cut

sub serialize : ActionClass('Serialize') {}
sub end :Private {
  my ($self, $c) = @_;
  $c->forward('serialize');
}


=head1 AUTHOR

David Santamauro

=head1 LICENSE

This library is free software. You can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

__PACKAGE__->config(
  'default'   => 'application/json',
  'stash_key' => 'rest',
  'map'       => {

         'text/xml'           => 'XML::Alexandria',
         'application/json'   => 'JSON::Alexandria',
         'text/x-json'        => 'JSON::Alexandria',

         'text/html'          => 'YAML::HTML',
         'text/x-data-dumper' => [ 'Data::Serializer', 'Data::Dumper' ],
         'text/x-data-denter' => [ 'Data::Serializer', 'Data::Denter' ],
         'text/x-data-taxi'   => [ 'Data::Serializer', 'Data::Taxi'   ],
         'application/x-storable'   => [ 'Data::Serializer', 'Storable' ],
         'application/x-freezethaw' => [ 'Data::Serializer', 'FreezeThaw' ],
         'text/x-config-general'    => [ 'Data::Serializer', 'Config::General' ],
         'text/x-php-serialization' => [ 'Data::Serializer', 'PHP::Serialization' ],
        }
);


__PACKAGE__->meta->make_immutable;

1;
