package Alexandria::CDWS::Controller::Custom::GIWS;
use Moose;
use namespace::autoclean;

use Time::HiRes qw(gettimeofday tv_interval);
use JSON;
use XML::LibXML;
use XML::XML2JSON;

use Data::Dumper;
use Alexandria::Tools::AppNumFormat;
use Alexandria::Tools::DocNumFormat;

BEGIN {extends 'Catalyst::Controller::REST'; }

=head1 NAME

Alexandria::CDWS::Controller::Custom::ACWS - custom text web service


=head1 DESCRIPTION

Alexandria::CDWS::Controller::Custom::ACWS takes either a document ucid
or an application ucid (or permutations thereof) and returns a tws_application
structure based on application filing number. This service is application-number centric.

=cut

#### utility
sub _xml2json {
  my ($self, $xml) = @_;
  my $xjson = XML::XML2JSON->new();
  return $xjson->convert( $xml );
}

#### parsing of XML
sub _fetch_invention_data {
  my ( $self, $c, $ucids ) = @_;
}

#### request methods
sub _post_params {
  my ( $self, $c ) = @_;
  my %params;

  if($c->req->content_type eq 'application/x-www-form-urlencoded') {
    my $s = $c->request->parameters;
    foreach my $key ( keys %{$s} ) {
      my $val = $s->{$key};
      $val =~ s/^\s+//;
      $val =~ s/\s+$//;
      $params{$key} = $val;
    }
  }
  else {
    my $data;
    if(ref($c->request->body) =~ /File/) {
      my $tmp = $c->request->body;
      $tmp->seek(0,0);
      $data = do { local $/; <$tmp> };
    }
    elsif($c->request->body) {
      $data = $c->request->body;
    }
    else {
      my @a = ();
      my $h = $c->request->parameters;
      foreach my $key ( keys %{$h} ) {
        if(ref($h->{$key}) =~ /ARRAY/) {
          foreach my $f ( @{$h->{$key}} ) {
            push(@a, "$key=$f");
          }
        }
        else { push(@a, "$key=$h->{$key}"); }
      }
      $data = join('&', @a);
    }
    my $str = uri_unescape($data);
    $str =~ s/^\s+//;
    $str =~ s/\s+$//;
    my @a = split(/\&/, $str);
    foreach my $e (@a) {
      my ($key, $val) = split(/=/, $e);
      $params{$key} = $val;
    }
  }
  return \%params;
}

sub _search {
  my ( $self, $c, $q, $solrparams ) = @_;
  my $user = $c->user;

  my $model = $c->model( $user->idx_model );

  if( !$solrparams->{qt} ) { $solrparams->{qt} = $model->{params}->{qt}; }

$c->log->debug( "SEARCH: $q" );

  my $res = $model->search( $q, $solrparams );
  my $json = JSON->new->utf8->convert_blessed;
  my $obj;
  eval {
    $obj = $json->decode( $res->raw_response->content );
  };
  if( $@ ) {
    $c->log->error( 'JSON error: ' . $@ );
    return undef;
  }
  return $obj;
}

sub _build_search {
  my ( $self, $c, $ucid, $endpoint ) = @_;

  return sprintf("( %s: %s )",
              $endpoint,
              $ucid
  );
}
sub _invention_from_fam_ex {
  my ( $self, $c, $ucid, $struct, $corresponding_ucids ) = @_;

#$c->log->debug( Dumper( $struct ) );
#$c->log->debug( Dumper( $corresponding_ucids ) );

  # we assume the best
  my %meta = ( iid => $ucid, lookup_status => 'found', confidence => 100,);
  my %invention = ();

  # the keys to {struct} are family IDs
  my @keys = keys %{ $struct };

  # if there is more than one family, found-duplicate
  if( scalar( @keys ) > 1 ) {
    $c->log->info( "multiple family-id(s) found, returning found-duplicate");
    $meta{lookup_status} = 'found-duplicate';
    return \%meta;
  }

  # pluck first family-id
  my $fam = shift( @keys );
  if( !$fam || $fam == -1 ) {
    $c->log->info( "no available family-id found, returning not-found");
    $meta{lookup_status} = 'not-found';
    return \%meta;
  }

  my ( $ctry, $dnum, $kind ) = split( /\-/, $ucid );
  my @countries = ( $ctry );
  if( $ctry ne 'EP') { push( @countries, 'EP' ); }
  if( $ctry ne 'WO') { push( @countries, 'WO' ); }
  my @q = ( sprintf( "( fam: %s ) AND ( pnctry: ( %s ) )", $fam, join(' OR ', @countries ) ) );

  # if the search of legal status surfaced any
  # corresponding documents, add them here in an
  # OR construct as well
  if( $corresponding_ucids && scalar( @{ $corresponding_ucids } ) ) {
    $invention{ origin } = 'EP';
    my $clause = join( ' ', @q );
    @q = ();
    push( @q, "( $clause )" );
    push( @q, "OR" );
    push( @q, sprintf( "( pn: ( %s ) )", join(' OR ', @{ $corresponding_ucids } ) ) );
  }

  # collapse into a single string for submission
  my $query = join(' ', @q );

  my %solrparams = (
    'fl' => 'ucid',
    'rows' => 100,
    'sort' => 'pnkind desc',
  );

  my $obj = $self->_search( $c, $query, \%solrparams );
#$c->log->debug( Dumper( $obj ) );
  if( $obj && $obj->{response} ) {
    my @ucids = ();
    foreach my $doc ( @{ $obj->{response}->{docs} } ) {
      push( @{ $invention{ publications } }, $doc->{ ucid } );
    }

    $invention{ family_id } = $fam;
    $meta{ invention } = \%invention;
  }
  return \%meta;
}

sub _invention_search {
  my ( $self, $c, $ucid, $q ) = @_;
  my $user = $c->user;

  my %solrparams = (
    'fl' => 'ucid,an,fam,pri',
    'rows' => 10,
  );

  my $obj = $self->_search( $c, $q, \%solrparams );
  my %family = ();
  my @corresponding_documents = ();
  my @ucids = ();
  if( $obj && $obj->{response} ) {
    foreach my $doc ( @{ $obj->{response}->{docs} } ) {
      my $woapp;
      if( $doc->{pri} ) {
        foreach my $pri ( @{ $doc->{pri} } ) {
          if( $pri =~ /\-W$/ ) {
            ( $woapp ) = ( $pri =~ /^(.*)\-W$/ );
          }
        }
      }
      push( @ucids, $doc->{ucid} );
      my ( $dctry, $dnum, $dkind ) = split( /\-/, $doc->{ucid} );
      my ( $actry, $anum, $akind ) = split( /\-/, $doc->{an} );
      push( @{ $family{ $doc->{fam} } },
          { pn => $doc->{ucid},
            pnctry => $dctry,
            pnnum => $dnum,
            pnkind => $dkind,
            an => $doc->{an},
            anctry => $actry,
            annum => $anum,
            ankind => $akind,
            has_pct_priority => ( $woapp ) ? 1 : 0,
            pct_app => $woapp,
          }
      );
    }

    foreach my $ucid ( @ucids ) {
      if( $ucid =~ /^EP.*?A(1|2)$/ ) {
        my $cdocs = $c->model( $user->rdb_model )->tws_ls_corresponding_documents( $ucid );
        if( $cdocs ) {
          foreach my $cdoc ( @{ $cdocs } ) {
            push( @corresponding_documents, $cdoc );
          }
        }
      }
    }
  }
  return $self->_invention_from_fam_ex( $c, $ucid, \%family, \@corresponding_documents );
}

sub _invention_from_dn {
  my ( $self, $c, $ucid ) = @_;
  return $self->_invention_search(
                 $c,
                 $ucid, 
                 $self->_build_search( $c, $ucid, 'pn' )
  );
}

sub _invention_from_an {
  my ( $self, $c, $ucid ) = @_;
  return $self->_invention_search(
                 $c,
                 $ucid, 
                 $self->_build_search( $c, $ucid, 'an' )
  );
}

sub _segment_raw_data {
  my ( $self, $c, $data ) = @_;
  my $str = $data;
  $str =~ s/\n/ /g;  # remove all new lines
  $str =~ s/\s+/ /g; # tighten up spaces
  $str =~ s/^\s+//;
  $str =~ s/\s+$//;
  # our only delimiter is the comma
  my @a = split( /,/, $str );
  return \@a;
}

sub _process_input_parameters {
  my ( $self, $c, $params, $endpoint ) = @_;
  return $self->_segment_raw_data( $c, $params->{id} );
}

=head1 METHODS

=cut

sub _process_request {
  my ( $self, $c, $params, $endpoint ) = @_;
  my $user = $c->user;

  my $t = [gettimeofday];
  my $data = $params->{id};
  my ($content_type) = grep { $self->{map}->{$_} } @{$c->request->accepted_content_types};
#$c->log->debug( Dumper( $params ) );

  return $self->status_bad_request(
    $c, message => 'insufficient parameters'
  ) if !$data;

  my $ids = $self->_process_input_parameters( $c, $params, $endpoint );
#$c->log->debug( Dumper( $ids ) );

  my @result = ();

  foreach my $ucid ( @{ $ids } ) {
    # both _application_from_dn and _application_from_an
    # return a \%meta hash structure
    if( $endpoint eq 'dn') {
      push( @result,
           $self->_invention_from_dn(
              $c, $ucid
           )
      );
    }
    elsif( $endpoint eq 'an' ) {
      push( @result,
           $self->_invention_from_an(
              $c, $ucid
           )
      );
    }
  }

#$c->log->debug( Dumper( \%result ) );

  my %ret = (
    status => 'success',
    time => sprintf( "%.3f", tv_interval( $t ) ),
    method => "custom/giws/$endpoint",
    content => \@result,
  );
  return $self->status_ok( $c, entity => \%ret );
}

### document number end-point
sub dn : Local : ActionClass('REST') { }
sub dn_GET  :Args(0){
  my ( $self, $c ) = @_;
  return $self->_process_request ($c, $c->request->parameters, 'dn' );
}

sub dn_POST :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_process_request( $c, $self->_post_params( $c ), 'dn' );
}

### application number end-point
sub an : Local : ActionClass('REST') { }
sub an_GET  :Args(0){
  my ( $self, $c ) = @_;
  return $self->_process_request ($c, $c->request->parameters, 'an' );
}

sub an_POST :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_process_request( $c, $self->_post_params( $c ), 'an' );
}

### anaqua-specific application number end-point
sub aqs : Local : ActionClass('REST') { }
sub aqs_GET  :Args(0){
  my ( $self, $c ) = @_;
  return $self->_process_request ($c, $c->request->parameters, 'aqs' );
}

sub aqs_POST :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_process_request( $c, $self->_post_params( $c ), 'aqs' );
}


=head2 end

=cut

sub serialize : ActionClass('Serialize') {}
sub end :Private {
  my ($self, $c) = @_;
  $c->forward('serialize');
}


=head1 AUTHOR

David Santamauro

=head1 LICENSE

This library is free software. You can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

__PACKAGE__->config(
  'default'   => 'application/json',
  'stash_key' => 'rest',
  'map'       => {

         'text/xml'           => 'XML::Alexandria',
         'application/json'   => 'JSON::Alexandria',
         'text/x-json'        => 'JSON::Alexandria',

         'text/html'          => 'YAML::HTML',
         'text/x-data-dumper' => [ 'Data::Serializer', 'Data::Dumper' ],
         'text/x-data-denter' => [ 'Data::Serializer', 'Data::Denter' ],
         'text/x-data-taxi'   => [ 'Data::Serializer', 'Data::Taxi'   ],
         'application/x-storable'   => [ 'Data::Serializer', 'Storable' ],
         'application/x-freezethaw' => [ 'Data::Serializer', 'FreezeThaw' ],
         'text/x-config-general'    => [ 'Data::Serializer', 'Config::General' ],
         'text/x-php-serialization' => [ 'Data::Serializer', 'PHP::Serialization' ],
        }
);


__PACKAGE__->meta->make_immutable;

1;
