package Alexandria::CDWS::Controller::Custom::TWS;
use Moose;
use namespace::autoclean;

use Time::HiRes qw(gettimeofday tv_interval);
use JSON;
use XML::LibXML;
use XML::XML2JSON;

use Data::Dumper;
use Alexandria::Tools::AppNumFormat;
use Alexandria::Tools::DocNumFormat;
use Unicode::Normalize;
use utf8;

BEGIN {extends 'Catalyst::Controller::REST'; }

=head1 NAME

Alexandria::CDWS::Controller::Custom::TWS - custom text web service


=head1 DESCRIPTION

Search Catalyst Controller.

=head1 METHODS

=cut

##### parsing methods
sub _should_parse {
  my ($self, $c, $fld) = @_;
  my $includes = $c->stash->{includes};
  my $ret = ( exists( $includes->{$fld} ) && $includes->{$fld} == 1 ) ? 1 : 0;
#$c->log->debug( "Should parse $fld: $ret" );
  return $ret;
}

sub _parse_patent_document {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;

  foreach my $patdoc ( $root->getElementsByTagName('patent-document') ) {
    my $fam = $patdoc->getAttribute('family-id');
    if( $fam && $fam != -1 ) { $output->{fam} = $fam; }
  }
}

sub _parse_publication_reference {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;

  foreach my $pubref ( $root->getElementsByTagName('publication-reference') ) {
    my $es = $pubref->getAttribute('entity-status');
    if( $es && !$output->{entity_status} ) {
      $output->{entity_status} = $es;
    }
    my %pubdata = (
      ucid => $pubref->getAttribute('ucid'),
      country => $pubref->findvalue('./document-id/country'),
      doc_number => $pubref->findvalue('./document-id/doc-number'),
      kind => $pubref->findvalue('./document-id/kind'),
      date => $pubref->findvalue('./document-id/date'),
      lang => $pubref->findvalue('./document-id/lang')
    );
    my $flag = $c->model('Custom')->app_or_grant( $pubdata{country}, $pubdata{kind}, $pubdata{date}, $pubdata{doc_number} );
#$c->log->debug( "$pubdata{ucid}: $flag" );
    if( $flag && $flag eq 'G' ) {
      # if it is a granted EP document of EP origin, it is the grant of tws
      if( $output->{origin} && $output->{origin} eq 'EP' && $pubdata{country} eq 'EP' ) {
        $output->{grant} = \%pubdata;
      }
      # if it is not an EP grant but origin is EP, it is the national-grant
      elsif( $output->{origin} && $output->{origin} eq 'EP' && $pubdata{country} ne 'EP' ) {
        $output->{national_grant} = \%pubdata;
      }
      # otherwise it is simply a grant
      else {
        $output->{grant} = \%pubdata
          if !$output->{grant};
      }
    }
    else {
      push( @{ $output->{publications} }, \%pubdata );
    }
  }
}

sub _parse_application_reference {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;

  foreach my $appref ( $root->getElementsByTagName('application-reference') ) {

    next if( !$appref->getAttribute('ucid') ||
      ( $output->{origin} && $output->{origin} eq 'EP' && $appref->getAttribute('ucid') !~ /^$self->{_input_country}/ ) );

    $output->{ucid} = $appref->getAttribute('ucid')
      if $appref->getAttribute('ucid') && !$output->{ucid};

    foreach my $docid ( $appref->getElementsByTagName('document-id') ) {
      my $format = $docid->getAttribute('format');
      # original vs epo
      if( $format && $format eq 'epo' ) {
        $output->{country} = $docid->findvalue('./country')
          if $docid->findvalue('./country') && !$output->{country};
        $output->{doc_number} = $docid->findvalue('./doc-number')
          if $docid->findvalue('./doc-number') && !$output->{doc_number};
        $output->{kind} = $docid->findvalue('./kind')
          if $docid->findvalue('./kind') && !$output->{kind};
        $output->{date} = $docid->findvalue('./date')
          if $docid->findvalue('./date') && !$output->{date};
        $output->{lang} = $docid->findvalue('./lang')
          if $docid->findvalue('./lang') && !$output->{lang};
        $output->{effective} = $output->{date} if !$output->{effective};
        $output->{format} = 'ifi';
#$c->log->debug( $appref->toString(1) );
      }
      else {
        $output->{doc_number_original} = $docid->findvalue('./doc-number')
          if $docid->findvalue('./doc-number') && !$output->{doc_number_original};
      }
    }
  }

  foreach my $pct ( $root->getElementsByTagName('pct-or-regional-filing-data') ) {
    if( $pct ) {
      $output->{origin} = 'WO' unless $output->{origin};
      my $pct_filed = $pct->findvalue('./document-id/date');
      if( $pct_filed && $pct_filed ne '' ) {
        # regardless of which is earlier, PCT filed trumps all other alternatives
        $output->{effective} = $pct_filed;
        $self->{_effective_is_pct} = 1;
      }
      # national-entry for US is filing date
#$c->log->debug( $pct->toString(1) );
      my $ned = $pct->findvalue('./national-entry-date/date');
      if( $ned ) {
        $output->{date} = $ned;
      }
    }
  }
}

sub _parse_priority {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;
  my $idx = 0;
  my @pr = ();
  my @pr_orig = ();

  # take first
  my $pris = $root->getElementsByTagName('priority-claims')->[0];
  if( $pris ) {
    foreach my $pri ( $pris->getElementsByTagName('priority-claim') ) {
      if ( $pri->getAttribute('load-source') eq 'docdb' ) {
        $idx++;
        push( @pr, {
            ucid => $pri->getAttribute('ucid'),
            country => $pri->findvalue('./document-id/country'),
            doc_number => $pri->findvalue('./document-id/doc-number'),
            kind => $pri->findvalue('./document-id/kind'),
            date => $pri->findvalue('./document-id/date'),
          }
        );
      }
      elsif ( $pri->getAttribute('load-source') eq 'patent-office' ) {
        $idx++;
        push( @pr_orig, {
            format => 'original',
            country => $pri->findvalue('./document-id/country'),
            doc_number => $pri->findvalue('./document-id/doc-number'),
            kind => $pri->findvalue('./document-id/kind'),
            date => $pri->findvalue('./document-id/date'),
          }
        );
      }
    }
  }
  if($idx == 0) { return; }

  my @sorted = sort {
    $a->{date} <=> $b->{date}
  } @pr;
  my @sorted_orig = sort {
    $a->{date} <=> $b->{date}
  } @pr_orig;

  while( my $earliest = shift( @sorted ) ) {
    # assign first one
    $output->{priority} = $earliest
      if !$output->{priority};
  }
  while( my $earliest = shift( @sorted_orig ) ) {
    # assign first one and clean up missing members
    if( !$earliest->{kind} ) { delete $earliest->{kind}; }
    if( !$earliest->{date} ) { delete $earliest->{date}; }
    $output->{priority_original} = $earliest
      if !$output->{priority_original};
    # add the original if we didn't get any docdb
    $output->{priority} = $earliest
      if !$output->{priority};
  }
}

sub _parse_term_extension {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;

  # take first ( any, really )
  my $pta = $root->getElementsByTagName('us-term-extension')->[0];
  if( $pta ) { $output->{pta} = $pta->textContent(); }
}

sub _parse_dates_of_public_availability {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;
  # <dates-of-public-availability>
  #   <registered-date><date>20040409</date></registered-date>
  # </dates-of-public-availability>
  if( my $dpa = $root->getElementsByTagName('dates-of-public-availability')->[0] ) {
    if( my $rd = $dpa->getElementsByTagName('registered-date')->[0] ) {
      $output->{effective_grant_date} = $rd->textContent();
    }
  }
}

sub _parse_titles {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;

  my %ttl = ();
  foreach my $ttl ( $root->getElementsByTagName('invention-title') ) {
    push( @{ $ttl{ $ttl->getAttribute('lang') } }, $ttl->textContent() );
  }

# $c->log->debug( Dumper( \%ttl ) );

  # now, pop off first of each language
  foreach my $lang ( sort keys %ttl ) {
    push( @{ $output->{titles} }, { lang => $lang, title => shift( @{$ttl{$lang}} ) } );
  }
}

sub _parse_related_documents {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;

  my %dates = (); # for effective date calculation
  my $reldocs = $root->getElementsByTagName('related-documents')->[0]; # take first
  if( $reldocs ) {
    foreach my $reldoc ( $reldocs->getElementsByTagName('relation') ) {
      if( $reldoc->getAttribute('type') =~ /(continuation|division)/) {
        if( !$output->{relation} ) {
          $output->{relation} = $reldoc->getAttribute('type');
          $output->{parent} = $reldoc->findvalue('./parent-doc/@ucid');
        }
        if( $reldoc->findvalue('./parent-doc/document-id/date') ) {
          $dates{ $reldoc->findvalue('./parent-doc/document-id/date') } = 1;
        }
      }
    }
    if( $output->{effective} ) {
      $dates{ $output->{effective} } = 1;
    }
    $dates{ $output->{date} } = 1;
    my @a = sort keys %dates;
#$c->log->debug( Dumper( \%dates ) );
    $output->{effective} = shift( @a )
      if !$self->{_effective_is_pct} ;
  }
}

sub _parse_ifi_integrated_content {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;

  # take first
  my $ifi = $root->getElementsByTagName('ifi-integrated-content')->[0];
  if( $ifi ) {
    my $claims = $ifi->getElementsByTagName('ifi-claims')->[0];
    my $pte = $ifi->getElementsByTagName('ifi-term-extension')->[0];
    my $stdname = $ifi->getElementsByTagName('ifi-standardized-name-current')->[0];

    my @claims_hierarchy = ();
    if( $claims ) {
      foreach my $clm ( $claims->getElementsByTagName('ifi-claim') ) {
        my @claim_refs = ();
        foreach my $ref ( $clm->getElementsByTagName('ifi-claim-ref') ) {
          push( @claim_refs, $ref->getAttribute('num') );
        }
        push(
          @claims_hierarchy,
          {
            num => $clm->getAttribute('num'),
            type => $clm->getAttribute('type'),
            refs => \@claim_refs
          }
        );
      }
    }
    my %ifidata = (
      claims_total => ( $claims ) ? $claims->getAttribute('total') : undef,
      claims_independent => ( $claims ) ? $claims->getAttribute('independent') : undef,
      claims_hierarchy => ( scalar( @claims_hierarchy ) ) ? \@claims_hierarchy : undef,
      pte => ( $pte ) ? $pte->textContent() : undef,
      assignee => ( $stdname ) ? $stdname->findvalue('./addressbook/name') : undef,
    );

    foreach my $psta ( $ifi->getElementsByTagName('ifi-patent-status-description') ) {
      my $status;
      if( $psta->textContent() =~ /^(Pend|Grant|Active)/ ) { $status = 'ALIVE'; }
      else { $status = 'LAPSED'; }
      # basically ... 
      if( $psta->getAttribute('country') && $self->{_input_country} eq $psta->getAttribute('country') ) {
$c->log->debug( sprintf('PATENT STATUS: country: %s / status: %s', $psta->getAttribute('country'), $psta->textContent() ) );
        $ifidata{patent_status} = $status;
      }
    }
    # if we didn't get status by country, grab first one
    my $psta = $ifi->getElementsByTagName('ifi-patent-status-description')->[0];
    if( !$ifidata{patent_status} && $psta ) {
      my $status;
      if( $psta->textContent() =~ /^(Pend|Grant|Active)/ ) { $status = 'ALIVE'; }
      else { $status = 'LAPSED'; }
      $ifidata{patent_status} = $status;
$c->log->debug( 'PATENT STATUS: ' . $psta->textContent() );
    }

#$c->log->debug( Dumper( $ifidata{claims_hierarchy} ) );
    foreach my $key ( sort keys %ifidata ) {
      if ($ifidata{$key} ) { $output->{$key} = $ifidata{$key}; }
    }
  }
}

sub _parse_parties {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;

  my $parties = $root->getElementsByTagName('parties')->[0] || return;

  if( !$output->{assignee} ) {
    # take first
    # if we don't have assignee already, pick:
    #   docdb, if not docdb then original
    my %as;
    foreach my $asg ( $parties->getElementsByTagName('assignee') ) {
      $as{ $asg->getAttribute('format') } =
        ( $asg->findvalue('./addressbook/name') )
            ? $asg->findvalue('./addressbook/name')
            : $asg->findvalue('./addressbook/last-name');
      last; # take first one
    }
#$c->log->debug( Dumper( \%as ) );
    if( $as{epo} )       { $output->{assignee} = $as{epo}; }
    elsif( $as{original} ) { $output->{assignee} = $as{original}; }
  
    if( !$output->{assignee} ) {
#$c->log->debug( $parties->toString(1) );
      # grab first docdb applicant
      foreach my $asg ( $parties->getElementsByTagName('applicant') ) {
        if( $asg->getAttribute('format') eq 'epo' ) {
          $output->{assignee} =
            ( $asg->findvalue('./addressbook/name') )
                ? $asg->findvalue('./addressbook/name')
                : $asg->findvalue('./addressbook/last-name');
          last;
        }
      }
    }
  }

  if( $self->_should_parse( $c, 'inventors') ) {
    my %in = ();
    foreach my $inv ( $parties->getElementsByTagName('inventor') ) {
      my $sequence = $inv->getAttribute('sequence');
#$c->log->debug( $inv->toString(1) );
      my $n = ( $inv->findvalue('./addressbook/name') )
              ? $inv->findvalue('./addressbook/name')
              : $inv->findvalue('./addressbook/last-name') . ', ' . $inv->findvalue('./addressbook/first-name');
      $n =~ s/,//g;
      $n =~ s/;//g;
      $n =~ s/\.//g;
      $n =~ s/^\s+//;
      $n =~ s/\s+$//;
      $in{ uc($n) } = $sequence;
    }
#$c->log->debug( Dumper( \%in ) );
    foreach my $n ( sort { $in{$a} <=> $in{$b} } keys %in ) {
      push( @{ $output->{inventors} }, $n );
    }
  }
}

sub _parse_classifications {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;

  # take first
  my $techical_data = $root->getElementsByTagName('technical-data')->[0];
  # grab all /sigh/
  # ic, uc, cpc, fi, fterms, ecla
  if( !$techical_data ) { return; }

  my %ipc;
  my %ipcr;
  my %cpc;
  my %nc;
  my %fi;
  my %ft;
  my %ecla;

  my $elements_a = {
    'classification-ipcr' => { type => 'ipcr', struct => \%ipcr },
    'classification-cpc' => { type => 'cpc', struct => \%cpc },
    'classification-ecla' => { type => 'ecla', struct => \%ecla },
    'f-term' => { type => 'ft', struct => \%ft },
  };
  # these have <main|further>
  my $elements_b = {
    'classification-ipc' => { type => 'ipc', struct => \%ipc },
    'classification-national' => { type => 'nc', struct => \%nc },
  };

  foreach my $key ( sort keys %{$elements_a} ) {
    my $name = $key;
    my $type = $elements_a->{$key}->{type};
    my $s = $elements_a->{$key}->{struct};
    
    foreach my $cls ( $techical_data->getElementsByTagName( $name ) ) {
      my $data = substr( $cls->textContent(), 0, 16 );
      $data =~ s/\s+$//;
      $s->{$data} = 1;
    }
    my @c = sort keys ( %{$s} );
    if( scalar( @c ) ) {
      push( @{ $output->{classifications} }, { $type => \@c } );
    }
  }

  foreach my $key ( sort keys %{$elements_b} ) {
    my $name = $key;
    my $type = $elements_b->{$key}->{type};
    my $s = $elements_b->{$key}->{struct};
    
    foreach my $contcls ( $techical_data->getElementsByTagName( $name ) ) {
      foreach my $sname ('main-classification', 'further-classification') {
        foreach my $cls ( $contcls->getElementsByTagName( $sname ) ) {
          my $data = substr( $cls->textContent(), 0, 16 );
          $data =~ s/\s+$//;
          $s->{$data} = 1;
        }
      }
    }
    my @c = sort keys ( %{$s} );
    if( scalar( @c ) ) {
      push( @{ $output->{classifications} }, { $type => \@c } );
    }
  }
#$c->log->debug( Dumper( $output->{classifications} ) );
}

sub _xml2json {
  my ($self, $xml) = @_;
  my $xjson = XML::XML2JSON->new();
  return $xjson->convert( $xml );
}

sub _parse_abstract {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;

  # original lang (assuming patent-document/@lang)
  my $lang = 'XX';
  my $pd = $root->getElementsByTagName('patent-document')->[0];
  if( $pd->getAttribute('lang') ) { $lang = $pd->getAttribute('lang'); }

  # english
  foreach my $ab ( $root->getElementsByTagName('abstract') ) {
    if( $ab->getAttribute('lang') =~ /en/i ) {
      $output->{abstract_en} = $ab->toString();
    }
    if( $ab->getAttribute('lang') =~ /$lang/i ) {
      $output->{abstract_original} = $ab->toString();
    }
  }
}

sub _parse_description {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;

  # original lang (assuming patent-document/@lang)
  my $lang = 'XX';
  my $pd = $root->getElementsByTagName('patent-document')->[0];
  if( $pd->getAttribute('lang') ) { $lang = $pd->getAttribute('lang'); }

  # english
  foreach my $desc ( $root->getElementsByTagName('description') ) {
    if( $desc->getAttribute('lang') =~ /en/i ) {
      $output->{description_en} = $desc->toString();
    }
    if( $desc->getAttribute('lang') =~ /$lang/i ) {
      $output->{description_original} = $desc->toString();
    }
  }
}

sub _parse_claims {
  my $self = shift;
  my $c = shift;
  my $root = shift || return;
  my $output = shift;

#$c->log->debug( "Parsing claims ...");

  # make no country assumptions
  my $ctry = 'XX';
  # original lang (assuming patent-document/@lang)
  my $lang = 'XX';
  my $pd = $root->getElementsByTagName('patent-document')->[0];
  if( $pd->getAttribute('country') ) { $ctry = $pd->getAttribute('country'); }
  if( $pd->getAttribute('lang') ) { $lang = $pd->getAttribute('lang'); }

  # don't overwrite IFI data (if available)
  if( !$output->{claims_total} ) {
    # take first language
    my $claims = $root->getElementsByTagName('claims')->[0];
    if( !$claims ) { return; }
    foreach my $claim ( $claims->getElementsByTagName('claim') ) {
      if( $ctry eq 'KR' ) {
        my $snippet = substr( $claim->textContent(), 0, 32 );
        $snippet =~ s/^\s+//;
        $snippet =~ s/\s+$//;
#$c->log->debug( "$ctry:$lang:$snippet" );
        if( $snippet =~ /^삭제/ || $snippet =~ /^delete/i ) {
          next;
        }
        else { $output->{claims_total}++; }
      }
      else {
        $output->{claims_total}++;
      }
    }
  }

  # english
  foreach my $claims ( $root->getElementsByTagName('claims') ) {
    if( $claims->getAttribute('lang') =~ /en/i ) {
#$c->log->debug( "Capturing claims_en ..." );
      $output->{claims_en} = $claims->toString();
    }
    if( $claims->getAttribute('lang') =~ /$lang/i ) {
#$c->log->debug( "Capturing claims_original ..." );
      $output->{claims_original} = $claims->toString();
    }
  }
#$c->log->debug( Dumper( $output->{claims_en} ) );
}

sub _parse_tws_object {
  my ($self, $c, $obj) = @_;
  my $user = $c->user;
  my %o = ();
  my $p = XML::LibXML->new( huge => 1 );
  my $dom;
  eval {
    $dom = $p->parse_string( $obj->{xml} );
  };
  if( $@ ) {
    $c->log->error( $@ );
    return undef;
  }
#delete $obj->{xml};
#$c->log->debug( Dumper( $obj ) );

  my $params = $c->request->parameters;
  if( $params->{origin} ) { $o{origin} = $params->{origin}; }

  # IMPORTANT: for multiple patent-documents, order is by kind desc
  # so B comes before A etc ...
  my $root = $dom->documentElement();

  # short-circuit non-matching applications (ACWS-52)
  if( $o{origin} && $o{origin} eq 'EP' && !$params->{'ep-not-found'} ) {
    my $found = 0;
    foreach my $appref ( $root->getElementsByTagName('application-reference') ) {
      next if( !$appref->getAttribute('ucid') || $appref->getAttribute('ucid') !~ /^$self->{_input_country}/ );
      $found = 1;
    }
    if( !$found ) { return undef; }
  }

  $self->_parse_patent_document( $c, $root, \%o );
  $self->_parse_publication_reference( $c, $root, \%o );
  $self->_parse_application_reference( $c, $root, \%o );
  $self->_parse_priority( $c, $root, \%o );
  $self->_parse_term_extension( $c, $root, \%o );
  $self->_parse_dates_of_public_availability( $c, $root, \%o );
  $self->_parse_titles( $c, $root, \%o );
  $self->_parse_related_documents( $c, $root, \%o );
  $self->_parse_ifi_integrated_content( $c, $root, \%o );
  $self->_parse_parties( $c, $root, \%o );
  if( $self->_should_parse( $c, 'classifications') ) {
    $self->_parse_classifications( $c, $root, \%o );
  }
  if( $self->_should_parse( $c, 'abstract_en')
        || $self->_should_parse( $c, 'abstract_original') ) {
    $self->_parse_abstract( $c, $root, \%o );
  }
  if( $self->_should_parse( $c, 'description_en')
        || $self->_should_parse( $c, 'description_original') ) {
    $self->_parse_description( $c, $root, \%o );
  }

  # slightly more complicated than it should be ...
  if( $self->_should_parse( $c, 'claims_en' )
        || $self->_should_parse( $c, 'claims_original')
        || ( $o{country} ne 'US' && $self->_should_parse( $c, 'claims_total') ) ) {
    $self->_parse_claims( $c, $root, \%o );
  }

  if( $self->_should_parse( $c, 'simple_family' ) ) {
    my $pd = $root->getElementsByTagName('patent-document')->[0];
    my $famobj = $c->model( $user->rdb_model )->family_simple( $pd->getAttribute('ucid') );
    $o{simple_family} = $famobj->{members};
  }
  if( $self->_should_parse( $c, 'extended_family' ) ) {
    my $pd = $root->getElementsByTagName('patent-document')->[0];
    my $ef = $c->model('MSSQL')->extended_inpadoc_family( $pd->getAttribute('ucid') );
    my @a = sort( @{ $ef } );
    $o{extended_family} = \@a;
  }

  ### FIX-UP
  if( $o{doc_number_original} && $o{country} eq 'US' ) {
    $o{doc_number_original} =~ s/^(..)/$1\//;
  }
  return \%o;
}

#### request methods
sub _post_params {
  my ( $self, $c ) = @_;
  my %params;

  if($c->req->content_type eq 'application/x-www-form-urlencoded') {
    my $s = $c->request->parameters;
    foreach my $key ( keys %{$s} ) {
      my $val = $s->{$key};
      $val =~ s/^\s+//;
      $val =~ s/\s+$//;
      $params{$key} = $val;
    }
  }
  else {
    my $data;
    if(ref($c->request->body) =~ /File/) {
      my $tmp = $c->request->body;
      $tmp->seek(0,0);
      $data = do { local $/; <$tmp> };
    }
    elsif($c->request->body) {
      $data = $c->request->body;
    }
    else {
      my @a = ();
      my $h = $c->request->parameters;
      foreach my $key ( keys %{$h} ) {
        if(ref($h->{$key}) =~ /ARRAY/) {
          foreach my $f ( @{$h->{$key}} ) {
            push(@a, "$key=$f");
          }
        }
        else { push(@a, "$key=$h->{$key}"); }
      }
      $data = join('&', @a);
    }
    my $str = uri_unescape($data);
    $str =~ s/^\s+//;
    $str =~ s/\s+$//;
    my @a = split(/\&/, $str);
    foreach my $e (@a) {
      my ($key, $val) = split(/=/, $e);
      $params{$key} = $val;
    }
  }
  return \%params;
}

sub _fetch_output {
  my ( $self, $c, $an, $docs ) = @_;
  my $user = $c->user;

  my $t = [gettimeofday];

  my ($actry, $anum, $akind) = split(/\-/, $docs->[0]->{an} );


  my @ucids = ();
  foreach my $doc ( @{ $docs } ) { push( @ucids, $doc->{ucid} ); }

#$c->log->debug( Dumper( \@ucids ) );

  my $includes = $c->stash->{includes};
  my $ftflag = 0;
  if( $includes->{abstract_en}
        || $includes->{abstract_original}
        || $includes->{claims_en}
        || $includes->{claims_original} ) { $ftflag = 1; }
  $ftflag = 1;
  return $self->_parse_tws_object(
    $c,
    $c->model( $user->rdb_model )->acws_container( \@ucids, $ftflag )
    # $c->model( $user->rdb_model )->tws_container( \@ucids, $ftflag )
  );
}

sub _meta_search_ex {
  my ( $self, $c, $q, $params ) = @_;
$c->log->debug("_meta_search_ex( $q )");
  my $user = $c->user;
  my $model = $c->model( $user->idx_model );
  my %solrparams = (
    'qt' => $model->{params}->{qt},
    'fl' => '*,score',
    'echoParams' => 'explicit',
  );
  if( $params ) {
    foreach my $key ( keys %{ $params } ) {
      $solrparams{ $key } = $params->{$key};
    }
  }
#$c->log->debug( Dumper( \%solrparams ) );
#  if( $famfrag ) { $solrparams{fl} = 'fam'; }

  my %ret = ( status => 'success');
  my $res = $model->search( "$q", \%solrparams );
  my $json = JSON->new->utf8->convert_blessed;
  my $obj;
  eval {
    $obj = $json->decode( $res->raw_response->content );
  };
  if( $@ ) {
    $ret{status} = 'error';
    $ret{message} = 'JSON error: failed to read response object';
    return \%ret;
  }
  $ret{content} = $obj;

  return \%ret;
}

sub _meta_search {
  my ( $self, $c, $input, $fl, $anfrag ) = @_;
  my $user = $c->user;
  my $model = $c->model( $user->idx_model );
  my %solrparams = (
    'qt' => $model->{params}->{qt},
    'fl' => '*,score',
    'echoParams' => 'explicit',
  );
  # for /dn enpoints, we only care about the
  # underlying application number(s)
  if( $anfrag ) { $solrparams{fl} = 'an'; }

  my %ret = ( status => 'success');
  my $res = $model->search( "$fl:$input", \%solrparams );
  my $json = JSON->new->utf8->convert_blessed;
  my $obj;
  eval {
    $obj = $json->decode( $res->raw_response->content );
  };
  if( $@ ) {
    $ret{status} = 'error';
    $ret{message} = 'JSON error: failed to read response object';
    return \%ret;
  }
  $ret{content} = $obj;

  return \%ret;
}

sub _tws_ls_correspondence {
  my ( $self, $c, $ucids ) = @_;
  my $user = $c->user;

  if( !$ucids || !scalar( @{ $ucids } ) ) { return undef; }

  foreach my $ucid ( @{ $ucids } ) {
    if( $ucid =~ /^EP.*?A(1|2|8|9)$/ ) {
      my $ret = $c->model( $user->rdb_model )->tws_ls_corresponding_documents( $ucid );
#$c->log->debug( Dumper( $ret ) );
      return $ret;
    }
  }
}

sub _origin_path {
  my ( $self, $c, $fmt ) = @_;
  my $input = $fmt->{input};

  my $meta;
  my $obj;
  my @records = ();
  my $numDocs;

  # short-circuit
  if( $fmt && $fmt->{origin} ) {
    # we lookup on {alt}, and re-submit family AND pnctry:
    my $a = $fmt->alt->{ucid};
    $a =~ s/\-XX$//;
    $meta = $self->_meta_search_ex( $c,
      sprintf("( anctry: ( %s OR %s ) AND annum: %s )",
           $fmt->{country}, $fmt->alt->{country}, $fmt->{app_number} )
    );
    #$meta = $self->_meta_search_ex( $c, "an:$a" );
    $obj = $meta->{content};
    $numDocs = $obj->{response}->{numFound};
    if( $numDocs > 0 ) {
      my $docs = $obj->{response}->{docs};
      my @cucids = ();
      foreach my $d ( @{ $docs } ) {
        push( @records, $d );
        push( @cucids, $d->{ucid} );
      }
$c->log->debug( Dumper( \@cucids, ) );

      my $cdocs = ($fmt->{country} eq 'DE' )
                     ? $self->_tws_ls_correspondence( $c, \@cucids )
                     : [];
$c->log->debug( Dumper( $cdocs ) );
#$c->log->debug( Dumper( $docs ) );
#$c->log->debug( "FAM: $docs->[0]->{fam}" );
      if( $docs->[0]->{fam} && $docs->[0]->{fam} != -1 ) {
        if( scalar( @{ $cdocs } ) ) {
          # we actually have a corresponding doc
          $meta = $self->_meta_search_ex( $c,
                      sprintf( "pn:( %s )", join(' OR ', @{ $cdocs } ) )
          );
        }
        else {
          $meta = $self->_meta_search_ex( $c,
                      sprintf( "pnctry: %s +fam: %d", $fmt->country, $docs->[0]->{fam} )
          );
        }
        $obj = $meta->{content};
        $numDocs = $obj->{response}->{numFound};
        # if( $numDocs == 0 ) { return undef; } # don't even have a family
$c->log->debug( "RET: " . Dumper( $obj->{responseHeader} ) );
        my %an_ucids = ();
        if( $numDocs > 0 ) {
          my $docs = $obj->{response}->{docs};
          foreach my $doc ( @{$docs} ) {
            my @a = split( '-',  $doc->{an} );
            $an_ucids{ $doc->{an} } = \@a;
          }
          my $_favor_only_a = 0;
          my $_favor_only_w = 0;
          if( scalar( keys %an_ucids ) > 1 ) {
            my %kinds = ();
            foreach my $ucid ( keys %an_ucids ) {
              $kinds{ $an_ucids{ $ucid }->[2] } = 1;
            }
$c->log->debug( "_origin_path::KINDS: " . Dumper( \%kinds ) );

            # basically, if there is an A and T kind
            # and origin = EP, we take the A
            if( $fmt->{origin} eq 'EP' ) {
              if( $kinds{A} && ( $kinds{T} || $kinds{W} ) ) { $_favor_only_a = 1; }
              else {
                $self->{_found_duplicate} = 1;
                $self->{_duplicate_ucids} = \%an_ucids;
              }
            }
            if( $fmt->{origin} eq 'WO' ) {
              if( $kinds{A} && $kinds{W} ) { $_favor_only_w = 1; }
              else {
                $self->{_found_duplicate} = 1;
                $self->{_duplicate_ucids} = \%an_ucids;
              }
            }
          }
          foreach my $doc ( @{$docs} ) {
            next if $_favor_only_a  && $doc->{an} !~ /A$/;
            next if $_favor_only_w  && $doc->{an} !~ /W$/;
            push( @records, $doc );
          }
        }
      }
#$c->log->debug( "_origin_path::RECORDS: " . Dumper( \@records ) );
      return $self->_fetch_output( $c, $input, \@records )
        if scalar( @records );
    }
  }
  return undef;
}
  
sub _an_data_ex {
  my ( $self, $c, $fmt ) = @_;
  my $input = $fmt->{input};

  if( my $obj = $self->_origin_path( $c, $fmt ) ) {
    return $obj;
  }

  ### otherwise, hack our way with US application numbers
  my $meta;
  my $obj;
  my @records = ();
  my $numDocs;

  # we have confidence if a date was submitted
  if( $c->request->parameters->{date} ) {
    my $a = $fmt->ucid;
    $a =~ s/\-XX$//;
    $meta = $self->_meta_search_ex( $c, "an: $a" );
    $obj = $meta->{content};
#$c->log->debug( Dumper( $obj ) );
    @records = ();
    $numDocs = $obj->{response}->{numFound};
  }
  else {
    # utterly RIDICULOUS!!!!
    #
    # if we don't use the date-results... we need to decide on a few things
    # 1) if $fmt->{series} >= 13 then we try with ????nnnnnn AND anseries
    # 2) if $fmt->{series < 13 then we try with nnnnnn?? AND anseries
    my @a = ();
    if( $fmt && $fmt->{series} >= 13) {
      @a = ( sprintf( "anctry: %s", $fmt->country ),
             sprintf( "annum: ????%02d%06d", $fmt->series, $fmt->app_number)
      );
    }
    else {
      my $anum = sprintf("%d", $fmt->app_number);
      @a = ( sprintf( "anctry: %s", $fmt->country ),
             sprintf( "annum: %d", $fmt->app_number)
      );
      my $t = ( 6 - length( $anum ) + 1 );
      if( $t == 1 ) { $a[1] .= '??'; }
      else { for(my $i = 0; $i < $t && $i < 2; $i++) { $a[1] .= '?'; } }
    }
    if( $fmt->{series} ) {
      push( @a, sprintf( "anseries: %02d", $fmt->{series} ) );
    }
    $meta = $self->_meta_search_ex( $c, join( " AND ", @a ) );
    $obj = $meta->{content};
    @records = ();
    $numDocs = $obj->{response}->{numFound};
  }

  # sub-hash of {an}
  my %an_ucids = ();
  if( $numDocs > 0 ) {
    my $docs = $obj->{response}->{docs};
    foreach my $doc ( @{$docs} ) {
      $an_ucids{ $doc->{an} } = 1;
      push( @records, $doc );
    }
    if( scalar( keys %an_ucids ) > 1 ) {
      $self->{_found_duplicate} = 1;
      $self->{_duplicate_ucids} = \%an_ucids;
$c->log->debug( "_an_date_ex::DUPS: " . Dumper(\%an_ucids) );
    }
    return $self->_fetch_output( $c, $input, \@records );
  }
  return undef;
}

sub _an_data {
  my ( $self, $c, $input, $fmt, $input_is_ucid ) = @_;

  if( my $obj = $self->_origin_path( $c, $fmt ) ) {
    return $obj;
  }
  else {
    if( $fmt->{origin} && $fmt->{origin} eq 'EP' ) {
      # ACWS-52: return not-found
      return undef;
    }
  }

  # am I trying to save time?
  # my $q = ( $input_is_ucid ) ? "( an:$input AND adyear:[1800 TO *] )" : "an:$input -an:$input" . 'D';
  my $q = "( an:$input AND adyear:[1800 TO *] )";
  my $meta = $self->_meta_search_ex( $c, $q );
  my $obj = $meta->{content};
#$c->log->debug( Dumper( $obj ) );
  my @records = ();
  my $numDocs = $obj->{response}->{numFound};

  if( $numDocs == 0 && $fmt && $fmt->alt ) {
    # try alternative
    my $ucid = $fmt->alt->{ucid};
    $ucid =~ s/\-XX$//;
    $meta = $self->_meta_search_ex( $c, "an:$ucid" );
    $obj = $meta->{content};
    $numDocs = $obj->{response}->{numFound};
  }

  if( $numDocs > 0 ) {
    my $docs = $obj->{response}->{docs};
    my %an_ucids = ();
    foreach my $doc ( @{$docs} ) {
      my @a = split('-',  $doc->{an} );
      $an_ucids{ $doc->{an} } = \@a;
    }
    my $_favor_only_a = 0;
    my $_favor_only_w = 0;
    if( scalar( keys %an_ucids ) > 1 ) {
      # NOTE NOTE NOTE: Code duplication -- see origin_path()
      my %kinds = ();
      foreach my $ucid ( keys %an_ucids ) {
        $kinds{ $an_ucids{ $ucid }->[2] } = 1;
      }
$c->log->debug( "_origin_path::KINDS: " . Dumper( \%kinds ) );

      # basically, if there is an A and T kind
      # and origin = EP, we take the A
      if( !$fmt->{origin} || $fmt->{origin} eq 'EP' ) {
        if( $kinds{A} && ( $kinds{T} || $kinds{W} ) ) { $_favor_only_a = 1; }
        else {
          $self->{_found_duplicate} = 1;
          $self->{_duplicate_ucids} = \%an_ucids;
        }
      }
      if( $fmt->{origin} && $fmt->{origin} eq 'WO' ) {
        if( $kinds{A} && $kinds{W} ) { $_favor_only_w = 1; }
        else {
          $self->{_found_duplicate} = 1;
          $self->{_duplicate_ucids} = \%an_ucids;
$c->log->debug( "(_an_data::DUPS: " . Dumper(\%an_ucids) );
        }
      }
    }
    foreach my $doc ( @{$docs} ) {
      next if $_favor_only_a  && $doc->{an} !~ /A$/;
      next if $_favor_only_w  && $doc->{an} !~ /W$/;
      push( @records, $doc );
    }
#$c->log->debug( Dumper( \@records ) );
    return $self->_fetch_output( $c, $input, \@records );
  }
  return undef;
}

### the kernel of information
# a tws_application One application number
sub _meta_from_an {
  my ( $self, $c, $input, $fmt ) = @_;

  my %meta = (
    iid => $fmt->input,
    lookup_status => 'not-found',
    confidence => 100
  );

  if( $fmt && ( $fmt->country() eq 'US' ) ) {
    if( my $and = $self->_an_data_ex( $c, $fmt ) ) {
      push( @{ $meta{tws_applications} }, $and );
    }
  }
  else {
    if( my $and = $self->_an_data( $c, $input, $fmt ) ) {
      push( @{ $meta{tws_applications} }, $and );
    }
  }
  if( $meta{tws_applications} ) {
    $meta{lookup_status} = "found";
    if( $self->{_found_duplicate} ) {
      $meta{lookup_status} = "found-duplicate";
      my @dups = sort keys %{ $self->{_duplicate_ucids} };
      $meta{duplicates} = \@dups;
    }
  }
  return \%meta;
}

# a document-number lookup can yield
# multiple tws_application(s).
#
# If:
#   - there are no results: not-found/100
#   - they all share the same application number: found/100
#   - the there are > 1 application numbers in the results (duplicate/100)
sub _meta_from_dn {
  my ( $self, $c, $input, $fmt ) = @_;
  my $sret = $self->_meta_search_ex( $c, "pn:$input", { fl => 'an' } );

  my %meta = ( iid => $fmt->input);
  if( $sret->{status} eq 'error' ) {
    return \%meta;
  }

  my $data = $sret->{content};
  my %an = ();
  foreach my $doc ( @{ $data->{response}->{docs} } ) {
    $an{ $doc->{an} } = 1;
  }

#$c->log->debug("_meta_from_dn: " . Dumper( \%an ) );

  # results could be
  #   0   { @lookup-status="not-found" }
  #   1   { @lookup-status="found" }
  #   > 1 { @lookup-status="duplicate" }
  my $n = scalar( keys %an );
  if( !$n ) {
    $meta{lookup_status} = "not-found";
    $meta{confidence} = 100;
  }
  elsif( $n == 1 ) {
    $meta{lookup_status} = "found";
    $meta{confidence} = 100;
    my @keys = keys %an;
    push( @{ $meta{tws_applications} },
       $self->_an_data( $c, shift(@keys), undef, 1 )
    );
  }
  else {
    $meta{lookup_status} = "found-duplicate";
    $meta{confidence} = 100;
    foreach my $anum ( keys %an ) {
      push( @{ $meta{tws_applications} },
         $self->_an_data( $c, $anum, undef, 1 )
      );
    }
  }
  return \%meta;
}

sub _segment_raw_data {
  my ( $self, $c, $data ) = @_;
  my $str = $data;
  $str =~ s/\n/ /g;  # remove all new lines
  $str =~ s/\s+/ /g; # tighten up spaces
  $str =~ s/^\s+//;
  $str =~ s/\s+$//;
  # the only delimiter is the comma
  my @a = split( /,/, $str );
  return \@a;
}

# valid parameters:
#   (ct) confidenceThreshold => <int>
#     TODO: define factors
#   (nf) notFound => [0|1]
#     Return list of input numbers that have no match
#   (dp) duplicates
#     Return only results where input has multiple matches
#   (ir) includeRequired => [0|1]
#     Inlcude required field (TODO: what is required field?)
#   (it) includeText => [0|1]
#     Include optional text fields (TODO: define optional vs required)
#   (if) includeField  => fl[,fl,fl]
#     Only out fields listed (TODO: define field names)
#   (ef) excludeField => fl[,fl,fl]
#     Suppress output of these fields
#   (dd) deduplicate
#     remove identical responses (TODO: define identical)
sub _handle_includes {
  my ( $self, $c ) = @_;
  my %required = (
    ucid => 1, country => 1, doc_number => 1, doc_number_original => 1, kind => 1, date => 1, lang => 1, format => 1, effective => 1, origin => 1, fam => 1,
    entity_status => 1,
    publications => 1, grant => 1,
    effective_grant_date => 1,
    national_grant => 1,
    pta => 1,
    terminal_disclaimer => 1, # TODO
    relation => 1, parent => 1,
    examination_request_date => 1, # TODO
    priority => 1,
    priority_original => 1,
    patent_status => 1,
    assignee => 1,
    titles => 1,
  );
  my %optional = (
    inventors => 0,
    simple_family => 0,
    extended_family => 0,
    classifications => 0,
    claims_total => 0,
    claims_hierarchy => 0,
    claims_en => 0,
    claims_original => 0,
    abstract_en => 0,
    abstract_original => 0,
    description_en => 0,
    description_original => 0,
  );
  my %includes = (
    %required
  );

  my $params = $c->request->parameters;
  if( $params->{'include-only'} ) {
    # special
    my %s = ( ucid => 1 );
    if( ref( $params->{'include-only'} ) =~ /ARRAY/) {
      foreach my $opt ( @{ $params->{'include-only'} } ) {
        $s{$opt} = 1;
      }
    }
    else {
      my @a = split( /,/, $params->{'include-only'} );
      foreach my $io ( @a ) {
        $s{ $io } = 1;
      }
    }
#$c->log->debug( Dumper( \%s ) );
    return \%s;
  }

  if( $params->{'include-optional'} ) {
    foreach my $opt ( keys %optional ) { $includes{$opt} = 1; }
  }

  if( $params->{'include'} ) {
    if( ref( $params->{'include'} ) =~ /ARRAY/) {
      foreach my $opt ( @{ $params->{'include'} } ) {
        $includes{$opt} = 1;
      }
    }
    else {
      $includes{ $params->{'include'} } = 1;
    }
  }
  if( $params->{'exclude'} ) {
    if( ref( $params->{'exclude'} ) =~ /ARRAY/) {
      foreach my $opt ( @{ $params->{'exclude'} } ) {
        delete $includes{$opt};
      }
    }
    else {
      delete $includes{ $params->{'exclude'} };
    }
  }
  return \%includes;
}

sub _process_input_parameters {
  my ( $self, $c, $params, $endpoint ) = @_;

  my @x = ();
  my $ids = $self->_segment_raw_data( $c, $params->{id} );
  foreach my $id ( @{ $ids } ) {
    my %h = ( id => $id );
    # If there is only origin/date without subscripts, they apply to all input
    if( $params->{origin} ) { $params->{origin} = uc( $params->{origin} ); }
    if( $params->{origin} && $params->{origin} =~ /^(EP|WO)/ ) { $h{origin} = $params->{origin}; }
    if( $params->{date} )   { $h{date} = $params->{date}; }
    push( @x, \%h );
  }

  foreach my $param ( keys %{ $params } ) {
    if( $param =~ /^origin\.(\d)$/ ) {
      my $sub = $1 - 1;
      $x[ $sub ]->{origin} = uc( $params->{ $param } );
    }
    elsif( $param =~ /^date\.(\d)$/ ) {
      my $sub = $1 - 1;
      $x[ $sub ]->{date} = $params->{ $param };
    }
  }

  foreach my $s ( @x ) {
    my %fmtparams = ( input => $s->{id} );
    if( $s->{date} )    { $fmtparams{date} = $s->{date}; }
    if( $s->{origin} )  { $fmtparams{origin} = $s->{origin}; }
    my $fmt;
    if( $endpoint eq 'aqs' )   { $s->{fmt} = Alexandria::Tools::AppNumFormat->new( \%fmtparams ); }
    elsif( $endpoint eq 'an' ) { $s->{fmt} = Alexandria::Tools::AppNumFormat->new( \%fmtparams ); }
    else                       { $s->{fmt} = Alexandria::Tools::DocNumFormat->new( \%fmtparams ); }
  }
  return \@x;
}

sub _process_request {
  my ( $self, $c, $params, $endpoint ) = @_;
  my $user = $c->user;


  my $data = $params->{id};
  my ($content_type) = grep { $self->{map}->{$_} } @{$c->request->accepted_content_types};

#$c->log->debug( "CONTENT-TYPE: $content_type") if $content_type;
$c->log->debug( Dumper( $params ) );

  return $self->status_bad_request(
    $c, message => 'insufficient parameters'
  ) if !$data;

  my $t = [gettimeofday];

  my $includes = $self->_handle_includes( $c );
  $c->stash( includes => $includes ); # save for use everywhere
#$c->log->debug( Dumper( $includes ) );

  my $is = $self->_process_input_parameters( $c, $params, $endpoint );
$c->log->debug( Dumper( $is ) );
  my %result = ();
  foreach my $in ( @{ $is } ) {

    $self->{_found_duplicate} = 0;
    $self->{_duplicate_ucids} = undef;
    $self->{_input_country} = $in->{fmt}->{country};

    my $ucid = $in->{fmt}->{ucid};
    if( $ucid =~ /XX$/ ) { $ucid =~ s/\-XX$//; }
    my $meta;
    if( $endpoint eq 'dn') {
      $meta = $self->_meta_from_dn( $c, $ucid, $in->{fmt} );
    }
    elsif( $endpoint eq 'an' ) {
      $meta = $self->_meta_from_an( $c, $ucid, $in->{fmt} );
    }
    #if( $params->{'not-found'} && $meta->{status} eq 'not-found' )
#$c->log->debug( Dumper( $meta->{duplicates} ) );
    $result{ $in->{fmt}->input } = $meta;
  }

  foreach my $key ( keys %result ) {
    # key = input id (iid)
    my $container = $result{$key};
    foreach my $tws_app ( @{ $container->{tws_applications} } ) {
      foreach my $fld ( keys %{ $tws_app } ) {
        if( !exists( $includes->{$fld} ) || !$includes->{$fld} ) {
          delete $tws_app->{$fld};
        }
        else {
          if( ( !$content_type && $fld =~ /abstract|claims_en|claims_original|description_en|description_original/ )
                   || ( $content_type && $content_type !~ /xml/
                                      && $fld =~ /abstract|claims_en|claims_original|description_en|description_original/ ) ) {
            eval {
              $c->log->debug( "Converting $fld to JSON ..." );
              $tws_app->{$fld} = $self->_xml2json( $tws_app->{$fld} );
            };
            if( $@ ) {
              $c->log->debug( Dumper( $tws_app->{$fld} ) );
              $c->log->error( $@ );
            }
          }
        }
      }
    }
# $c->log->debug( Dumper( $container ) );
    if( !scalar( @{ $container->{tws_applications} } ) || $container->{lookup_status} eq 'found-duplicate' ) {
      delete $container->{tws_applications};
    }
  }
#$c->log->debug( Dumper( \%result ) );

  my %ret = (
    status => 'success',
    time => sprintf( "%.3f", tv_interval( $t ) ),
    method => "custom/tws/$endpoint",
    content => \%result,
  );
  return $self->status_ok( $c, entity => \%ret );
}

### document number end-point
sub dn : Local : ActionClass('REST') { }
sub dn_GET  :Args(0){
  my ( $self, $c ) = @_;
  return $self->_process_request ($c, $c->request->parameters, 'dn' );
}

sub dn_POST :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_process_request( $c, $self->_post_params( $c ), 'dn' );
}

### application number end-point
sub an : Local : ActionClass('REST') { }
sub an_GET  :Args(0){
  my ( $self, $c ) = @_;
  return $self->_process_request ($c, $c->request->parameters, 'an' );
}

sub an_POST :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_process_request( $c, $self->_post_params( $c ), 'an' );
}

### anaqua-specific application number end-point
sub aqs : Local : ActionClass('REST') { }
sub aqs_GET  :Args(0){
  my ( $self, $c ) = @_;
  return $self->_process_request ($c, $c->request->parameters, 'aqs' );
}

sub aqs_POST :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_process_request( $c, $self->_post_params( $c ), 'aqs' );
}


=head2 end

=cut

sub serialize : ActionClass('Serialize') {}
sub end :Private {
  my ($self, $c) = @_;
  $c->forward('serialize');
}


=head1 AUTHOR

David Santamauro

=head1 LICENSE

This library is free software. You can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

__PACKAGE__->config(
  'default'   => 'application/json',
  'stash_key' => 'rest',
  'map'       => {

         'text/xml'           => 'XML::Alexandria::XResponse::Custom::TWS',
         'application/json'   => 'JSON::Alexandria',
         'text/x-json'        => 'JSON::Alexandria',

         'text/html'          => 'YAML::HTML',
         'text/x-data-dumper' => [ 'Data::Serializer', 'Data::Dumper' ],
         'text/x-data-denter' => [ 'Data::Serializer', 'Data::Denter' ],
         'text/x-data-taxi'   => [ 'Data::Serializer', 'Data::Taxi'   ],
         'application/x-storable'   => [ 'Data::Serializer', 'Storable' ],
         'application/x-freezethaw' => [ 'Data::Serializer', 'FreezeThaw' ],
         'text/x-config-general'    => [ 'Data::Serializer', 'Config::General' ],
         'text/x-php-serialization' => [ 'Data::Serializer', 'PHP::Serialization' ],
        }
);


__PACKAGE__->meta->make_immutable;

1;
