package Alexandria::CDWS::Controller::Family::Meta;
use Moose;
use namespace::autoclean;

use Time::HiRes qw(gettimeofday tv_interval);
use JSON;
use XML::LibXML;
use XML::XML2JSON;

use Data::Dumper;
use Alexandria::Tools::AppNumFormat;
use Alexandria::Tools::DocNumFormat;
use Unicode::Normalize;
use utf8;

BEGIN {extends 'Catalyst::Controller::REST'; }

=head1 NAME

Alexandria::CDWS::Controller::Family::Meta - family meta informatin service


=head1 DESCRIPTION

Search Catalyst Controller.

=head1 METHODS

=cut

#### request methods
sub _post_params {
  my ( $self, $c ) = @_;
  my %params;

  if($c->req->content_type eq 'application/x-www-form-urlencoded') {
    my $s = $c->request->parameters;
    foreach my $key ( keys %{$s} ) {
      my $val = $s->{$key};
      $val =~ s/^\s+//;
      $val =~ s/\s+$//;
      $params{$key} = $val;
    }
  }
  else {
    my $data;
    if(ref($c->request->body) =~ /File/) {
      my $tmp = $c->request->body;
      $tmp->seek(0,0);
      $data = do { local $/; <$tmp> };
    }
    elsif($c->request->body) {
      $data = $c->request->body;
    }
    else {
      my @a = ();
      my $h = $c->request->parameters;
      foreach my $key ( keys %{$h} ) {
        if(ref($h->{$key}) =~ /ARRAY/) {
          foreach my $f ( @{$h->{$key}} ) {
            push(@a, "$key=$f");
          }
        }
        else { push(@a, "$key=$h->{$key}"); }
      }
      $data = join('&', @a);
    }
    my $str = uri_unescape($data);
    $str =~ s/^\s+//;
    $str =~ s/\s+$//;
    my @a = split(/\&/, $str);
    foreach my $e (@a) {
      my ($key, $val) = split(/=/, $e);
      $params{$key} = $val;
    }
  }
  return \%params;
}

sub _segment_raw_data {
  my ( $self, $c, $data ) = @_;
  my $str = $data;
  $str =~ s/\n/ /g;  # remove all new lines
  $str =~ s/\s+/ /g; # tighten up spaces
  $str =~ s/^\s+//;
  $str =~ s/\s+$//;
  # the only delimiter is the comma
  my @a = split( /,/, $str );
  return \@a;
}

### 
# one application number
sub _meta_from_an {
  my ( $self, $c, $input ) = @_;
  my %meta = ();
  return \%meta;
}

# a document-number
sub _meta_from_dn {
  my ( $self, $c, $ucid ) = @_;
  my $user = $c->user;

  my %meta = ( );

  my $family = $c->model( $user->rdb_model  )->family_meta( $ucid );

  $meta{ucid} = $ucid;
  foreach my $member ( @{$family->{members}} ) {
    $meta{family_id} = $member->{family_id} if !$meta{family_id};

    my @keys = qw( ucid published anucid filed title ifi_status epo_legal_status );
    my %h = ();
    foreach my $key ( @keys ) {
      if( $member->{$key} ) { $h{ $key } = $member->{ $key}; }
    }
    push( @{ $meta{members} },  \%h );
  }
  return \%meta;
}

sub _process_request {
  my ( $self, $c, $params, $endpoint ) = @_;
  my $user = $c->user;

  my $data = $params->{ucid};
  my ($content_type) = grep { $self->{map}->{$_} } @{$c->request->accepted_content_types};

# $c->log->debug( "CONTENT-TYPE: $content_type") if $content_type;
# $c->log->debug( Dumper( $params ) );

  return $self->status_bad_request(
    $c, message => 'insufficient parameters'
  ) if !$data;

  my $t = [gettimeofday];

  my $ucids = $self->_segment_raw_data( $c, $data );
$c->log->debug( Dumper( $ucids ) );

  my @families = ();

  foreach my $ucid ( @{$ucids} ) {

    my $meta;
    if( $endpoint eq 'dn') {
      $meta = $self->_meta_from_dn( $c, $ucid );
    }
    elsif( $endpoint eq 'an' ) {
      $meta = $self->_meta_from_an( $c, $ucid );
    }
    push( @families, $meta );
  }

  my %ret = (
    status => 'success',
    time => sprintf( "%.3f", tv_interval( $t ) ),
    method => "family/meta/$endpoint",
    content => \@families
  );
  return $self->status_ok( $c, entity => \%ret );
}

### document number end-point
sub dn : Local : ActionClass('REST') { }
sub dn_GET  :Args(0){
  my ( $self, $c ) = @_;
  return $self->_process_request ($c, $c->request->parameters, 'dn' );
}

sub dn_POST :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_process_request( $c, $self->_post_params( $c ), 'dn' );
}

### application number end-point
sub an : Local : ActionClass('REST') { }
sub an_GET  :Args(0){
  my ( $self, $c ) = @_;
  return $self->_process_request ($c, $c->request->parameters, 'an' );
}

sub an_POST :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_process_request( $c, $self->_post_params( $c ), 'an' );
}

=head2 end

=cut

sub serialize : ActionClass('Serialize') {}
sub end :Private {
  my ($self, $c) = @_;
  $c->forward('serialize');
}


=head1 AUTHOR

David Santamauro

=head1 LICENSE

This library is free software. You can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

__PACKAGE__->config(
  'default'   => 'application/json',
  'stash_key' => 'rest',
  'map'       => {

         'text/xml'           => 'XML::Alexandria::XResponse::Custom::TWS',
         'application/json'   => 'JSON::Alexandria',
         'text/x-json'        => 'JSON::Alexandria',

         'text/html'          => 'YAML::HTML',
         'text/x-data-dumper' => [ 'Data::Serializer', 'Data::Dumper' ],
         'text/x-data-denter' => [ 'Data::Serializer', 'Data::Denter' ],
         'text/x-data-taxi'   => [ 'Data::Serializer', 'Data::Taxi'   ],
         'application/x-storable'   => [ 'Data::Serializer', 'Storable' ],
         'application/x-freezethaw' => [ 'Data::Serializer', 'FreezeThaw' ],
         'text/x-config-general'    => [ 'Data::Serializer', 'Config::General' ],
         'text/x-php-serialization' => [ 'Data::Serializer', 'PHP::Serialization' ],
        }
);


__PACKAGE__->meta->make_immutable;

1;
