package Alexandria::CDWS::Controller::Report;
use Moose;
use namespace::autoclean;
use File::Path;
use Time::HiRes qw( gettimeofday tv_interval );

BEGIN {extends 'Catalyst::Controller::REST'; }


=head1 NAME

Alexandria::Controller::Report - main reporting controller


=head1 DESCRIPTION

Catalyst Controller.

=head1 METHODS

=cut

use Data::Dumper;

sub columns : Local : ActionClass('REST') { }
sub columns_GET  :Args(0){
  my ( $self, $c ) = @_;

  my $t = [gettimeofday];

  my $ret = $c->model('Report')->columns();

  $ret->{time} = sprintf(" %.6f", tv_interval($t) );

#$c->log->debug( Dumper($ret) );
  return $self->status_ok( $c, entity => $ret );
}

sub listex : Local : ActionClass('REST') { }
sub listex_GET  :Args(0){
  my ( $self, $c ) = @_;

  my $t = [gettimeofday];

  my $params = $c->request->parameters;

  my $ret = $c->model('Report')->listex( $params );

  $ret->{time} = sprintf( "%.6f", tv_interval($t) );
#  $c->log->debug( Dumper($ret) );
  return $self->status_ok( $c, entity => $ret );
}

sub list : Local : ActionClass('REST') { }
sub list_GET  :Args(0){
  my ( $self, $c ) = @_;

  my $t = [gettimeofday];

  my $params = $c->request->parameters;

  my $ret = $c->model('Report')->list( $params );

  $ret->{time} = sprintf( "%.6f", tv_interval($t) );
#  $c->log->debug( Dumper($ret) );
  return $self->status_ok( $c, entity => $ret );
}

sub _handle_create_params {
  my ($self, $c, $params) = @_;

  if($params->{type} eq 'csv') {
    #### CSV
    my %opts = (
      # defaults
      'csv.cols' => '__all__',
      'csv.header' => 1,
      'csv.nonl' => 1,
      'csv.stubs' => 1,
      'csv.maxcolsize' => -1,
      'csv.incbasicinextended' => 0,
    );
    foreach my $key ( sort keys %{$params} ) {
      if($key =~ /^csv\./) {
        $opts{$key} = $params->{$key};
        delete $params->{$key};
      }
    }
    my @a = ();
    foreach my $key ( keys %opts ) {
      push(@a, "$key=$opts{$key}");
    }
    if(scalar(@a)) {
      $params->{csvopts} = join('&', @a);
    }
  }

  if($params->{source} eq 'search') {
    my $user = $c->user;

    # SOLR parameters are set here (internally)
    $params->{solrurl} = $c->model( $user->idx_model )->{server};

#$c->log->debug("SOLR Params: $params->{solrparams}");
    if($params->{solrparams}) {
      my %h;
      my @a = split( /\~/, $params->{solrparams} );
      foreach my $pair ( @a ) {
        my ( $key, $val ) = split( /=/, $pair );
        if( $val ) { $h{$key} = $val; }
      }
      if( !$h{qt} && $c->model( $user->idx_model )->{params}->{qt} ) {
        $h{qt} = $c->model( $user->idx_model )->{params}->{qt};
      }
      $params->{solrparams} = join( "&", map { "$_=$h{$_}" } keys %h);
    }
    else {
      $params->{solrparams} = sprintf( "qt=%s", $c->model( $user->idx_model )->{params}->{qt} );
    }
#$c->log->debug("SOLR Params: $params->{solrparams}");
  }
  elsif($params->{source} eq 'list') {
  }
  elsif($params->{source} eq 'table') {
  }
#$c->log->debug("Params: " . Dumper( $params ) );
  return $params;
}

sub _handle_queue_params {
  my ($self, $c, $params) = @_;

  # some defaults
  $params->{package_type} ||= 'single';
  $params->{package_compression} ||= 'zip';
  return $params;
}

sub create : Local : ActionClass('REST') { }
sub create_GET  :Args(0){
  my ( $self, $c ) = @_;

  my $t = [gettimeofday];

  my $params = $c->request->parameters;
#$c->log->debug( Dumper($params) );

  my $ret = $c->model('Report')->create(
    $self->_handle_create_params( $c, $params )
  );

  $ret->{time} = sprintf( "%.6f", tv_interval($t) );

  return $self->status_ok( $c, entity => $ret );
}

sub create_POST  :Args(0){
  my ( $self, $c ) = @_;
  my %params;

  my $t = [gettimeofday];

  if($c->req->content_type eq 'application/x-www-form-urlencoded') {
    my $s = $c->request->parameters;
    foreach my $key ( keys %{$s} ) {
      my $val = $s->{$key};
      $val =~ s/^\s+//;
      $val =~ s/\s+$//;
      $params{$key} = $val;
    }
  }
  else {
    my $data;
    if(ref($c->request->body) =~ /File/) {
      my $tmp = $c->request->body;
      $tmp->seek(0,0);
      $data = do { local $/; <$tmp> };
    }
    elsif($c->request->body) {
      $data = $c->request->body;
    }
    else {
      $data = $c->request->parameters;
    }
    my $str = uri_unescape($data->{qp});
    $str =~ s/^\s+//;
    $str =~ s/\s+$//;
    my @a = split(/\&/, $str);
    foreach my $e (@a) {
      my ($key, $val) = split(/=/, $e);
      $params{$key} = $val;
    }
  }

  # force json
  $c->req->content_type( 'application/json' );
  my $ret = $c->model('Report')->create(
    $self->_handle_create_params( $c, \%params )
  );
  $ret->{time} = sprintf( "%.6f", tv_interval($t) );

#  $c->log->debug( Dumper($ret) );
  return $self->status_ok( $c, entity => $ret );
}

sub detail : Local : ActionClass('REST') { }
sub detail_GET  :Args(0){
  my ( $self, $c ) = @_;

  my $t = [gettimeofday];
  my $params = $c->request->parameters;

  if(!$params->{id}) {
    return $self->status_bad_request( $c, message => 'missing required id parameter' );
  }
  $params->{tracker_id} = $params->{id};

  my $ret = $c->model('Report')->detail( $params );
  $ret->{time} = sprintf(" %.6f", tv_interval($t) );

#  $c->log->debug( Dumper($ret) );
  return $self->status_ok( $c, entity => $ret );
}

sub queue : Local : ActionClass('REST') { }
sub queue_GET  :Args(0){
  my ( $self, $c ) = @_;

  my $t = [gettimeofday];

  my $params = $c->request->parameters;

  if(!$params->{id}) {
    return $self->status_bad_request( $c, message => 'missing required id parameter' );
  }
  $params->{tracker_id} = $params->{id};

  my $ret = $c->model('Report')->queue(
    $self->_handle_queue_params($c, $params)
  );
  $ret->{time} = sprintf( "%.6f", tv_interval($t) );

#  $c->log->debug( Dumper($ret) );
  return $self->status_ok( $c, entity => $ret );
}

sub status : Local : ActionClass('REST') { }
sub status_GET  :Args(0){
  my ( $self, $c ) = @_;

  my $t = [gettimeofday];

  my $params = $c->request->parameters;

  if(!$params->{id}) {
    return $self->status_bad_request( $c, message => 'missing required id parameter' );
  }
  $params->{tracker_id} = $params->{id};

  my $ret = $c->model('Report')->status( $params );
  if( $ret->{report}->{running_status} eq 'complete'
       && $ret->{report}->{completed_status} eq 'success'
       && $ret->{report}->{completed_url} ) {
    $ret->{report}->{completed_url} =
             sprintf("%s", $c->uri_for( $c->model('Report')->{alias} . $ret->{report}->{completed_url} ) );

    if($params->{redirect}) {
      $c->redirect( $ret->{report}->{completed_url} );
    }
  }
  $ret->{time} = sprintf(" %.6f", tv_interval($t) );

#$c->log->debug( Dumper($ret) );
  return $self->status_ok( $c, entity => $ret );
}

sub delete : Local : ActionClass('REST') { }
sub delete_GET  :Args(0){
  my ( $self, $c ) = @_;

  my $t = [gettimeofday];

  my $params = $c->request->parameters;
  my $storage = $c->model('Report')->{storage};

  if(!$params->{id}) {
    return $self->status_bad_request( $c, message => 'missing required id parameter' );
  }
  $params->{tracker_id} = $params->{id};

  my $ret = $c->model('Report')->delete( $params );
  if($ret->{status} eq 'success') {
    # remove from storage ...
    if(!$storage || !(-W "$storage")) {
      $ret->{storage} = { status => 'error', message => 'storage unavailable' };
    }
    my $path = "$storage/$params->{id}";
    if(!rmdir("$path")) {
      $ret->{storage} = { status => 'error', message => $! };
    }
  }
  $ret->{time} = sprintf(" %.6f", tv_interval($t) );

# $c->log->debug( Dumper($ret) );
  return $self->status_ok( $c, entity => $ret );
}

=head2 end

=cut

sub serialize : ActionClass('Serialize') {}
sub end :Private {
  my ($self, $c) = @_;
  $c->forward('serialize');
}


=head1 AUTHOR

David Santamauro

=head1 LICENSE

This library is free software. You can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

__PACKAGE__->config(
  'default'   => 'application/json',
  'stash_key' => 'rest',
  'map'       => {

         'text/xml'           => 'XML::Simple',
         'application/json'   => 'JSON::Alexandria',
         'text/x-json'        => 'JSON::Alexandria',

         'text/html'          => 'YAML::HTML',
         'text/x-data-dumper' => [ 'Data::Serializer', 'Data::Dumper' ],
         'text/x-data-denter' => [ 'Data::Serializer', 'Data::Denter' ],
         'text/x-data-taxi'   => [ 'Data::Serializer', 'Data::Taxi'   ],
         'application/x-storable'   => [ 'Data::Serializer', 'Storable' ],
         'application/x-freezethaw' => [ 'Data::Serializer', 'FreezeThaw' ],
         'text/x-config-general'    => [ 'Data::Serializer', 'Config::General' ],
         'text/x-php-serialization' => [ 'Data::Serializer', 'PHP::Serialization' ],
        }
);


__PACKAGE__->meta->make_immutable;

1;
