package Alexandria::CDWS::Controller::Report::Validate;
use Moose;
use namespace::autoclean;
use File::Path;
use Time::HiRes qw( gettimeofday tv_interval );

BEGIN {extends 'Catalyst::Controller::REST'; }


=head1 NAME

Alexandria::Controller::Report::Validate - validate input numbers


=head1 DESCRIPTION

Catalyst Controller.

=head1 METHODS

=cut

use Data::Dumper;

sub _parse_array {
  my $self = shift;

  return $_[0]  if(ref $_[0]);
  return [ @_ ] if(@_ > 1);

  my $val = $_[0];

  return undef  unless(defined $val);

  $val =~ s/^ (?:\[.+\]=)? \{ (.*) \} $/$1/sx;

  my @array;

  while($val =~ s/(?:"((?:[^"\\]+|\\.)*)"|([^",]+))(?:,|$)//) {
    my($item) = map { $_ eq 'NULL' ? undef : $_ } (defined $1 ? $1 : $2);
    $item =~ s{\\(.)}{$1}g  if(defined $item);
    push(@array, $item);
  }
  return \@array;
}

sub _parse_suggestions {
  my $self = shift;
  my $data = shift;
  my @ret = ();
  
  my $a = $self->_parse_array($data);
  foreach my $suggestion (@{$a}) {
    $suggestion =~ s/^\((.*)\)$/$1/;
    
    my %s = ();
    my @a = split(',', $suggestion);
    $s{publication_id} = $a[0];
    $s{ucid} = $a[1];
    $s{published} = $a[2];
    $s{title} = $a[3];
    
    if($s{title}) { $s{title} =~ s/\"//g; }

    push(@ret, \%s);
  }
  return \@ret;
}

sub _build_search_us_app {
  my ( $self, $c, $fmt, $input_is_ucid ) = @_;

  if( $input_is_ucid ) {
    return "( an: $fmt->{ucid} )";
  }

  if( $c->request->parameters->{date} ) {
    my $a = $fmt->ucid;
    $a =~ s/\-XX$//;
    return "an:$a";
  }
  else {
    my @a = ();
    if( $fmt && $fmt->{series} && $fmt->{series} >= 13) {
      @a = ( sprintf( "anctry: %s", $fmt->country ),
             sprintf( "annum: ????%02d%06d", $fmt->series, $fmt->app_number)
      );
    }
    else {
      my $anum = sprintf("%d", $fmt->app_number);
      @a = ( sprintf( "anctry: %s", $fmt->country ),
             sprintf( "annum: %d", $fmt->app_number)
      );
      my $t = ( 6 - length( $anum ) + 1 );
      if( $t == 1 ) { $a[1] .= '??'; }
      else { for(my $i = 0; $i < $t && $i < 2; $i++) { $a[1] .= '?'; } }
    }
    if( $fmt->{series} ) {
      push( @a, sprintf( "anseries: %02d", $fmt->{series} ) );
    }
    return sprintf("%s AND ( ( adyear: [1700 TO *] ) )", join( ' AND ', @a ) );
  }
}

sub _build_search {
  my ( $self, $c, $fmt, $endpoint, $input_is_ucid ) = @_;
  if( $fmt->{country} eq 'US' && $endpoint eq 'an' ) {
    return $self->_build_search_us_app( $c, $fmt, $input_is_ucid );
  }
  if( $fmt && $fmt->{alt} ) {
    return sprintf("( %s: (%s OR %s) ) AND ( ( %s: %s ) ) AND ( ( adyear: [1700 TO *] ) )",
               ( $endpoint eq 'an' ) ? 'anctry' : 'pnctry',
                   $fmt->{country}, $fmt->{alt}->{country},
               ( $endpoint eq 'an' ) ? 'annum'  : 'pnnum',
               ( $endpoint eq 'an' ) ?  $fmt->{app_number} : $fmt->{doc_number}
    );
  }
  else {
    my $ucid = $fmt->{ucid};
    $ucid =~ s/\-XX$//i;
    return sprintf("( ( %s: %s ) AND ( adyear: [1700 TO *] ) ) %s",
                $endpoint,
                $ucid,
                ( $endpoint eq 'an' ) ? 'NOT ( ankind:W )' : ''
    );
  }
}

sub _search {
  my ( $self, $c, $input ) = @_;
  my $params = $c->request->parameters;

  my $t = [gettimeofday];

  my @validations = ();

  foreach my $entry (split(/\n/, $input)) {
    $entry =~ s/^\s+//;
    $entry =~ s/\s+$//;
    if(!$entry) { next; } # skip blank lines ...

    my $fmt;
    my %fmtargs = ( input => $entry );
    if( $params->{origin} ) { $fmtargs{origin} = $params->{origin}; }
    if( $params->{date} )   { $fmtargs{date} = $params->{date}; }
    $fmt = Alexandria::Tools::AppNumFormat->new( %fmtargs );
    my $q = $self->_build_search( $c, $fmt, 'an' );
    push( @validations, { ucid => $fmt->{ucid}, q => $q } );
  }

  my %content = (
    status => 'success',
    time => sprintf('%.3f', tv_interval( $t ) ),
    content => {
      total => 0,
      matched => 0,
      duplicates => 0,
      validations => \@validations
    }
  );

#$c->log->debug( Dumper( \%content ) );

  return $self->status_ok( $c, entity => \%content );
}


sub _post_params {
  my ( $self, $c ) = @_;
  my %params;

  if($c->req->content_type eq 'application/x-www-form-urlencoded') {
$c->log->debug( __PACKAGE__ . ':: content-type: ' . $c->request->content_type );
    my $s = $c->request->parameters;
    foreach my $key ( keys %{$s} ) {
      my $val = $s->{$key};
      $val =~ s/^\s+//;
      $val =~ s/\s+$//;
      $params{$key} = $val;
    }
  }
  else {
    my $data;
    if(ref($c->request->body) =~ /File/) {
$c->log->debug( __PACKAGE__ . ':: FileRef' );
      my $tmp = $c->request->body;
      $tmp->seek(0,0);
      $data = do { local $/; <$tmp> };
    }
    elsif($c->request->body) {
$c->log->debug( __PACKAGE__ . ':: RequestBody' );
      $data = $c->request->body;
    }
    else {
$c->log->debug( __PACKAGE__ . ':: RequestParameters' );
      my @a = ();
      my $h = $c->request->parameters;
      foreach my $key ( keys %{$h} ) {
        if(ref($h->{$key}) =~ /ARRAY/) {
          foreach my $f ( @{$h->{$key}} ) {
            push(@a, "$key=$f");
          }
        }
        else { push(@a, "$key=$h->{$key}"); }
      }
      $data = join('&', @a);
    }
$c->log->debug( __PACKAGE__ . ":: DATA: $data" );
    my $str = uri_unescape($data);
    $str =~ s/^\s+//;
    $str =~ s/\s+$//;
    my @a = split(/\&/, $str);
    foreach my $e (@a) {
      my ($key, $val) = split(/=/, $e);
      $params{$key} = $val;
    }
  }
  return \%params;
}

sub _do_init {
  my ( $self, $c, $params, $endpoint ) = @_;

  my $user = $c->user;
  my $t = [gettimeofday];

  my @validations = ();

  my $input = $params->{input};
  $input =~ s/\r\n/,/g;
  $input =~ s/\s+/ /g;
  $input =~ s/,/\n/g;
  
  if( $endpoint eq 'an' || $params->{is_application} ) {
    return $self->_search( $c, $input );
  }

  # this belongs in the Model ... 
  my $model = $c->model( $user->rdb_model );
  my $dbh = $model->dbh();

  # one validation per user ...
  $dbh->do( 'delete from reports.t_report_validations where user_id=' . $user->id );

  my $sth = $dbh->prepare(" insert into reports.t_report_validations
                                ( format, user_id, user_input, input_type, formatted_input )
                                values (?,?,?,?,?)");

  foreach my $entry (split(/\n/, $input)) {
    $entry =~ s/^\s+//;
    $entry =~ s/\s+$//;
    if(!$entry) { next; } # skip blank lines ...

    my $fmt;
    my %fmtargs = ( input => $entry );
    if( $params->{origin} ) { $fmtargs{origin} = $params->{origin}; }
    if( $params->{format} ) { $fmtargs{hint} = $params->{format}; }
    if( $params->{date} )   { $fmtargs{date} = $params->{date}; }
    $fmt = Alexandria::Tools::DocNumFormat->new( %fmtargs );
    $sth->execute( $params->{format}, $user->id, $fmt->input, $endpoint, $fmt->ucid );
  }

  $sth->finish();

  $sth = $dbh->prepare( "select * from reports.t_report_validations where user_id=? order by input_id");
  $sth->execute( $user->id );

  my ( $total, $matched, $dups ) = ( 0, 0, 0 );
  while( my $ref = $sth->fetchrow_hashref() ) {
    $total++;
    if( $ref->{matched} == 1 )   { $matched++; }
    if( $ref->{duplicate} == 1 ) { $dups++; }
    $ref->{suggestions} = $self->_parse_suggestions( $ref->{suggestions} );
    foreach my $key ( keys %{$ref} ) {
      delete $ref->{$key} if !$ref->{$key};
    }

    $ref->{ucid} = $ref->{formatted_input}
      if( !$ref->{suggestions} || !scalar( @{$ref->{suggestions} } ) );
    push( @validations, $ref );
  }
  $sth->finish();

  my %content = (
    status => 'success',
    time => sprintf('%.3f', tv_interval( $t ) ),
    content => {
      total => $total,
      matched => $matched,
      duplicates => $dups,
      validations => \@validations
    }
  );

#$c->log->debug( Dumper( \%content ) );

  return $self->status_ok( $c, entity => \%content );
}

sub init : Local : ActionClass('REST') { }
sub init_GET :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_do_init( $c, $c->request->parameters, 'dn' );
}
sub init_POST :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_do_init( $c, $self->_post_params( $c ), 'dn' );
}

sub dn : Local : ActionClass('REST') { }
sub dn_GET :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_do_init( $c, $c->request->parameters, 'dn' );
}
sub dn_POST :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_do_init( $c, $self->_post_params( $c ), 'dn' );
}

sub an : Local : ActionClass('REST') { }
sub an_GET :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_do_init( $c, $c->request->parameters, 'an' );
}
sub an_POST :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_do_init( $c, $self->_post_params( $c ), 'an' );
}

sub _do_update {
  my ( $self, $c, $params ) = @_;
  my $t = [gettimeofday];
  my %content = (
    status => 'success',
    time => sprintf('%.3f', tv_interval( $t ) ),
    params => $params
  );
  return $self->status_ok( $c, entity => $params );
}
sub update : Local : ActionClass('REST') { }
sub update_GET :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_do_update( $c, $c->request->parameters );
}
sub update_POST :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_do_update( $c, $self->_post_params( $c ) );
}


=head2 end

=cut

sub serialize : ActionClass('Serialize') {}
sub end :Private {
  my ($self, $c) = @_;
  $c->forward('serialize');
}


=head1 AUTHOR

David Santamauro

=head1 LICENSE

This library is free software. You can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

__PACKAGE__->config(
  'default'   => 'application/json',
  'stash_key' => 'rest',
  'map'       => {

         'text/xml'           => 'XML::Simple',
         'application/json'   => 'JSON::Alexandria',
         'text/x-json'        => 'JSON::Alexandria',

         'text/html'          => 'YAML::HTML',
         'text/x-data-dumper' => [ 'Data::Serializer', 'Data::Dumper' ],
         'text/x-data-denter' => [ 'Data::Serializer', 'Data::Denter' ],
         'text/x-data-taxi'   => [ 'Data::Serializer', 'Data::Taxi'   ],
         'application/x-storable'   => [ 'Data::Serializer', 'Storable' ],
         'application/x-freezethaw' => [ 'Data::Serializer', 'FreezeThaw' ],
         'text/x-config-general'    => [ 'Data::Serializer', 'Config::General' ],
         'text/x-php-serialization' => [ 'Data::Serializer', 'PHP::Serialization' ],
        }
);


__PACKAGE__->meta->make_immutable;

1;
