package Alexandria::CDWS::Controller::Root;
use Moose;
use namespace::autoclean;
use Data::Dumper;
use Time::HiRes qw( gettimeofday tv_interval );

BEGIN { extends 'Catalyst::Controller' }

# Sets the actions in this controller to be registered with no prefix
# so they function identically to actions created in MyApp.pm
#
__PACKAGE__->config(namespace => '');

=head1 NAME

Alexandria::MMC::Controller::Root - Root Controller for Alexandria MMC

=head1 DESCRIPTION

This module controls the root (/) of the Alexandria CDWS application. Its main
purpose is to provide default behavior as shown in the methods handled.

=head1 METHODS

=head2 auto

Note that 'auto' runs after 'begin' but before your actions

=cut
use Data::Dumper;
use JSON;
use Sys::Hostname;

sub _finalize {
  my ($c, $event, $args) = @_;

  $c->stash( __request_end => [gettimeofday] );

  my $user = $c->user;
  my ( $start, $end ) = ( $c->stash->{__request_begin}, $c->stash->{__request_end} );

  my $t = localtime;

  # imager is special
  # imager/w-300-h-300/KR/20161021/A/020160/12/20/73/pat00010.jpg
  my %data = (
    source_server => hostname,
    name => $user->name,
    address => $c->request->address,
    method => $c->request->method,
    path => ( $c->request->path =~ /^imager/ ) ? 'imager' : $c->request->path,
    parameters => $c->stash->{__parameters_original},
    response_code => $c->response->status,
    response_length => length($c->response->body) || 0,
    user_agent => $c->request->user_agent || '<no-user-agent>',
    request_time => sprintf('%.6f', tv_interval( $start, $end ) )
  );

  $c->model('Users')->log_statistics( \%data )
    if( $c->config->{log_statistics_database} );

  my @la = (
    $data{address},
    $data{name},
    sprintf("%s %s", $data{method}, $data{path}),
    $data{parameters},
    $data{response_code},
    $data{response_length},
    $data{source_server},
    $data{user_agent},
    $data{request_time}
  );
  $c->log->info( join("\t", @la ) )
    if( $c->config->{log_statistics_error_log} );
}

sub auto :Private {
  my ($self, $c) = @_;
  my $params = $c->request->parameters;

  $c->stash( __request_begin => [gettimeofday] );

  my $json = JSON->new();
  $json->utf8(1);
  $json->pretty(0);
  $c->stash( __parameters_original => $json->encode( $params ) );

  # Not all of these values will be here
  my %authinfo = ( user_name  => ( $c->request->headers->header('X-User') )  ?
                                      $c->request->headers->header('X-User') : '' ,
                   passwd     => ( $c->request->headers->header('X-Password') ) ?
                                      $c->request->headers->header('X-Password') : '',
                   ip_address => $c->request->address );

  my $aret = $c->authenticate(\%authinfo);
  if( ! $aret ) {
    $c->log->debug("authentication failed");
    $c->response->code(401);
    $c->response->body('Authentication Failed');
    $c->detach();
  }

  # store statistics at end of request
  $c->add_subscriber( 'finalize', \&_finalize );

  # the return value here is a
  #   Alexandria::CDWS::Authentication::Store::DBI::User
  #    { _store, _user and (internally), auth_realm is added }
  $c->stash(user => $c->user);

  return 1;
}

sub default :Path {
  my ($self, $c) = @_;
  $c->response->code(404);
}

=head1 AUTHOR

David Santamauro, E<lt>david.santamauro@fairviewresearch.comE<gt>

=head1 LICENSE

This library is free software. You can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

__PACKAGE__->config();


__PACKAGE__->meta->make_immutable;

1;

