package Alexandria::CDWS::Controller::Search;
use Moose;
use namespace::autoclean;

use JSON ();
use XML::LibXML;
use Data::Dumper;
use URI::Escape;
use Time::HiRes qw( gettimeofday tv_interval );

BEGIN {extends 'Catalyst::Controller::REST'; }

# parameter for search we need to sanitize
has 'start' =>
  ( is => 'rw', isa => 'Int', default => 0 );
has 'rows' =>
  ( is => 'rw', isa => 'Int', default => 10 );
has 'sort' =>
  ( is => 'rw', isa => 'Str', default => 'pd desc' );
has 'timeAllowed' =>
  ( is => 'rw', isa => 'Int', default => 300000 );

=head1 NAME

Alexandria::Controller::Search - Main search controller


=head1 DESCRIPTION

Search Catalyst Controller.

=head1 METHODS

=cut

use Data::Dumper;

sub schema : Local : ActionClass('REST') { }
sub schema_GET  :Args(0){
  my ( $self, $c ) = @_;

  my $t = [gettimeofday];

  my $user = $c->user;
  my $params = $c->request->parameters;
  my ($content_type) = grep { $self->{map}->{$_} } @{$c->request->accepted_content_types};

  if( !$content_type ) {
    $content_type = 'application/json';
  }

  if( $content_type =~ /xml/ ) { $params->{wt} = 'xml'; }
  else { $params->{wt} = 'json'; }

  my $path = $c->request->path;
  $path =~ s/^search\///;
  $path =~ s/\?.*$//;
  $params->{indent} = 'on';
  my $response = $c->model( $user->idx_model )->schema( $path, $params );
#$c->log->debug( Dumper($response->content) );
#$c->log->debug( Dumper($response->raw_response->content) );

  my %ret = (
    status => 'success',
    time => sprintf( "%.6f", tv_interval($t) ),
    content => $response->content
  );
  my $json = JSON->new->utf8->convert_blessed;
  $json->pretty(1);
  return $self->status_ok( $c, entity => $json->encode( \%ret ) );
}

sub _fix_start {
  my ($self, $model, $start) = @_;

  if( !$start || ( $start && $start > $model->{start_limit}) ) {
    return 0;
  }
  return $start;
}

sub _fix_timeAllowed {
  my ($self, $model,  $ta) = @_;

  if( !$ta || ( $ta && $ta > $model->{time_allowed} ) ) {
    $ta = $model->{time_allowed};
  }
  return $ta;
}

sub _sanitize_parameters {
  my ($self, $c, $params) = @_;
  my $user = $c->user;

  my $srchmodel = $c->model( $user->idx_model );
  my $rdbmodel = $c->model( $user->idx_model );

  my %solropts = (
    q => $params->{q},
    fq => $params->{fq},
    start => $self->_fix_start( $srchmodel, $params->{start} ),
    timeAllowed => $self->_fix_timeAllowed ( $srchmodel, $params->{timeAllowed} ),
    qt =>
      ( $params->{qt} &&
         ( $params->{qt} eq 'attachments' || $params->{qt} eq 'pitr' ) ) ? $params->{qt} : $srchmodel->{params}->{qt},
  );

  # delete all we have 'fixed'
  delete $params->{q};
  delete $params->{fq};
  delete $params->{start};
  delete $params->{timeAllowed};
  delete $params->{qt};

  # faceting moreLikeThis clustering
  foreach my $key ('facet', 'mlt', 'clustering') {
    if( $params->{$key} && $params->{$key} !~ /false/i) { $params->{$key} = 'true'; }
  }

  # indent output
  $params->{indent} = 'true';

#  # if we are index version 2.1, we stuff the collection parameter into
#  # the options
#  if( $c->config->{index_version} > 20 ) {
#    $solropts{collection} = $srchmodel->{params}->{collection}
#      if $srchmodel->{params}->{collection};
#    if( $srchmodel->{params}->{fq} ) {
#      my @fq = ( $srchmodel->{params}->{fq} );
#      if( $solropts{fq} ) {
#        push( @fq, $solropts{fq} );
#      }
#      $solropts{fq} = \@fq;
#    }
#  }

  # allow the rest to pass through
  foreach my $param ( keys %{$params} ) {
    $solropts{$param} = $params->{$param};
  }
  return \%solropts;
}

sub _build_final_response {
  my ($self, $c, $start_t, $response) = @_;

  # result ...
  my %ret = ();
  if( !$response || !$response->raw_response->content ) {
    $ret{status} = 'error';
    $ret{message} = 'Unknown error: SOLR response empty';
    return \%ret;
  }

  my $ct = $response->raw_response->header('content-type');
#$c->log->debug( "RESPONSE CONTENT TYPE: $ct" );
  if( $ct =~ /xml/) {
    $ret{xml} = $response->raw_response->content;
    return \%ret;
  }

  # result pager ...
  my %p = ();

  # convert to an obj for manipulate data ...
  my $json = JSON->new->utf8->convert_blessed;
  my $obj;
  eval {
    $obj = $json->decode( $response->raw_response->content );
  };
  if( $@ ) {
    $ret{status} = 'error';
    $ret{message} = 'JSON error: failed to read response object';
    return \%ret;
  }

  if( $obj && $obj->{responseHeader}->{status} == 0 ) {
    my $pager = $response->pager;
    if($pager) {
      $p{totalEntries} = $pager->total_entries;
      $p{entriesPerPage} = $pager->entries_per_page;
      $p{currentPage} = $pager->current_page;
      $p{entriesOnThisPage} = $pager->entries_on_this_page;
      $p{firstPage} = $pager->first_page;
      $p{lastPage} = $pager->last_page;
      $p{previousPage} = $pager->previous_page;
      $p{nextPage} = $pager->next_page;
      $obj->{responseHeader}->{pager} = \%p;
    }
  }

  # special parameter for attachment lists
  # ALPA-2612
  my $params = $c->request->parameters;
  my $user = $c->user;
  if( $obj && $obj->{response} && $obj->{response}->{docs} ) {
    foreach my $doc ( @{ $obj->{response}->{docs} } ) {
      if( $doc->{ucid} ) {
        if( $params->{'attachments'} ) {
          my $attachments = $c->model('Attachments')->list( $doc->{ucid} );
          if( scalar( @{ $attachments } ) ) {
            $doc->{attachments} = $attachments;
          }
        }
        if( $params->{'family.simple'} ) {
          my $family = $c->model( $user->rdb_model  )->family_simple( $doc->{ucid} );
          if( $family ) {
            $doc->{'family_simple'} = $family;
          }
        }
        if( $params->{'family.extended'} ) {
          my $family = $c->model( $user->rdb_model  )->family_extended(
                 $doc->{ucid},
                 $c->model( $user->idx_model ),
                 1, # include basic
                 $c->model( 'MSSQL' ), # mssql
                 1 # mssql-only
          );
          if( $family ) {
            $doc->{'family_extended'} = $family;
          }
        }
      }
    }
  }

  $ret{status} = 'success';
  $ret{time} = sprintf( "%.06f", tv_interval($start_t) );
  $ret{content} = $obj;
  return \%ret;
}

sub _handle_search {
  my ( $self, $c, $params ) = @_;

  my $t = [gettimeofday];

  my $user = $c->user;

  my $srchmodel = $c->model( $user->idx_model );
  my $rdbmodel = $c->model( $user->idx_model );

  my ($content_type) = grep { $self->{map}->{$_} } @{$c->request->accepted_content_types};

#$c->log->debug( __PACKAGE__ . ":: content-type: $content_type" );
#$c->log->debug( Dumper( $params ) );

  if(!$params->{q}) {
    return $self->status_bad_request($c, message => 'insufficient parameters');
  }

  # set timeout
  $srchmodel->agent->timeout(600);

  # clean up paramters
  my $solropts = $self->_sanitize_parameters( $c, $params );

  # allow the wt parameter
  if(!$solropts->{wt}) {
    $solropts->{wt} = 'json';
    if($content_type && $content_type =~ /x\-solr/) {
      $solropts->{wt} = 'xml';
    }
  }

$c->log->debug( Dumper ($solropts ) );

  # save locally and remove q
  my $q = $solropts->{q}; delete $solropts->{q};

  # fake out attachments and pitr ...
  if( $solropts->{qt} && $solropts->{qt} eq 'attachments' ) {
    $srchmodel = $c->model('Solr::Attachments');
    delete $solropts->{qt};
  }
  if( $solropts->{qt} && $solropts->{qt} eq 'pitr' ) {
    $srchmodel = $c->model('Solr::PITR');
    delete $solropts->{qt};
  }

  ### finally, the search ...
  my $response = $self->_build_final_response(
    $c,
    $t,
    $srchmodel->search( $q, $solropts )
  );

  # intercept x-solr
  if($content_type && $content_type =~ /x\-solr/) {
    # hash with one member: xml
    return $self->status_ok( $c, entity => $response );
  }

#$c->log->debug( Dumper ($response ) );

  my $ret;
  if($content_type && $content_type =~ /xml/) {
    my $xjson = XML::XML2JSON->new();

    $ret = { search => $response } ;
    
    #$ret = $xjson->obj2xml( { search => $response } );
  }
  else {
    my $json = JSON->new->utf8->convert_blessed;
    $ret = $json->pretty->encode( $response );
  }
  return $self->status_ok( $c, entity => $ret );
}

sub query : Local : ActionClass('REST') { }
sub query_GET  :Args(0){
  my ( $self, $c ) = @_;
  return $self->_handle_search( $c, $c->request->parameters );
}

sub _post_params {
  my ( $self, $c ) = @_;
  my %params;

#$c->log->debug( Dumper( $c->request ) );
  if($c->req->content_type eq 'application/x-www-form-urlencoded') {
$c->log->debug( __PACKAGE__ . ':: content-type: ' . $c->request->content_type );
    my $s = $c->request->parameters;
    foreach my $key ( keys %{$s} ) {
      my $val = $s->{$key};
      $val =~ s/^\s+//;
      $val =~ s/\s+$//;
      $params{$key} = $val;
    }
  }
  else {
    my $data;
    if(ref($c->request->body) =~ /File/) {
$c->log->debug( __PACKAGE__ . ':: FileRef' );
      my $tmp = $c->request->body;
      $tmp->seek(0,0);
      $data = do { local $/; <$tmp> };
    }
    elsif($c->request->body) {
$c->log->debug( __PACKAGE__ . ':: RequestBody' );
      $data = $c->request->body;
    }
    else {
$c->log->debug( __PACKAGE__ . ':: RequestParameters' );
      my @a = ();
      my $h = $c->request->parameters;
      foreach my $key ( keys %{$h} ) {
        if(ref($h->{$key}) =~ /ARRAY/) {
          foreach my $f ( @{$h->{$key}} ) {
            push(@a, "$key=$f");
          }
        }
        else { push(@a, "$key=$h->{$key}"); }
      }
      $data = join('&', @a);
    }
#$c->log->debug( __PACKAGE__ . ":: DATA: $data" );
    my $str = uri_unescape($data);
    $str =~ s/^\s+//;
    $str =~ s/\s+$//;
    my @a = split(/\&/, $str);
    foreach my $e (@a) {
      my ($key, $val) = split(/=/, $e);
      $params{$key} = $val;
    }
  }
  return \%params;
}

sub query_POST  :Args(0){
  my ( $self, $c ) = @_;
  return $self->_handle_search( $c, $self->_post_params( $c ) );
}

=head2 end

=cut

sub serialize : ActionClass('Serialize') {}
sub end :Private {
  my ($self, $c) = @_;
  # CAUTION: unknown side affects. Somehow, a POST
  # request, regardless of content-type and response->code,
  # sets the body to "Content-Type xxx is not supported"
  # For successul request/response, undef this because the
  # serializer sets the body
  if( $c->response->code == 200 && $c->response->has_body ) {
    $c->response->body(undef);
  }
  $c->forward('serialize');
}


=head1 AUTHOR

David Santamauro

=head1 LICENSE

This library is free software. You can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

__PACKAGE__->config(
  'default'   => 'application/json',
  'stash_key' => 'rest',
  'map'       => {

         'text/xml'           => 'XML::Alexandria',
         'application/json'   => 'JSON::Passthrough',
         'text/x-json'        => 'JSON::Passthrough',
         'text/x-solr'        => 'XML::Alexandria::XResponse::XSolr',

         'text/html'          => 'YAML::HTML',
         'text/x-data-dumper' => [ 'Data::Serializer', 'Data::Dumper' ],
         'text/x-data-denter' => [ 'Data::Serializer', 'Data::Denter' ],
         'text/x-data-taxi'   => [ 'Data::Serializer', 'Data::Taxi'   ],
         'application/x-storable'   => [ 'Data::Serializer', 'Storable' ],
         'application/x-freezethaw' => [ 'Data::Serializer', 'FreezeThaw' ],
         'text/x-config-general'    => [ 'Data::Serializer', 'Config::General' ],
         'text/x-php-serialization' => [ 'Data::Serializer', 'PHP::Serialization' ],
        }
);


__PACKAGE__->meta->make_immutable;

1;
