package Alexandria::CDWS::Controller::Session;
use Moose;
use namespace::autoclean;

use Time::HiRes qw(gettimeofday tv_interval );
use Data::Dumper;

BEGIN {extends 'Catalyst::Controller::REST'; }

=head1 NAME

Alexandria::Controller::Session - Main session controller


=head1 DESCRIPTION

Search Catalyst Controller.

=head1 METHODS

=cut

use Data::Dumper;

sub _post_params {
  my ( $self, $c ) = @_;
  my %params;

  if($c->req->content_type eq 'application/x-www-form-urlencoded') {
$c->log->debug( __PACKAGE__ . ':: content-type: ' . $c->request->content_type );
    my $s = $c->request->parameters;
    foreach my $key ( keys %{$s} ) {
      my $val = $s->{$key};
      $val =~ s/^\s+//;
      $val =~ s/\s+$//;
      $params{$key} = $val;
    }
  }
  else {
    my $data;
    if(ref($c->request->body) =~ /File/) {
$c->log->debug( __PACKAGE__ . ':: FileRef' );
      my $tmp = $c->request->body;
      $tmp->seek(0,0);
      $data = do { local $/; <$tmp> };
    }
    elsif($c->request->body) {
$c->log->debug( __PACKAGE__ . ':: RequestBody' );
      $data = $c->request->body;
    }
    else {
$c->log->debug( __PACKAGE__ . ':: RequestParameters' );
      my @a = ();
      my $h = $c->request->parameters;
      foreach my $key ( keys %{$h} ) {
        if(ref($h->{$key}) =~ /ARRAY/) {
          foreach my $f ( @{$h->{$key}} ) {
            push(@a, "$key=$f");
          }
        }
        else { push(@a, "$key=$h->{$key}"); }
      }
      $data = join('&', @a);
    }
    my $str = uri_unescape($data);
    $str =~ s/^\s+//;
    $str =~ s/\s+$//;
    my @a = split(/\&/, $str);
    foreach my $e (@a) {
      my ($key, $val) = split(/=/, $e);
      $params{$key} = $val;
    }
  }
  return \%params;
}

sub list : Local : ActionClass('REST') { }
sub list_GET  :Args(0){
  my ( $self, $c ) = @_;

  my $t = [gettimeofday];
  my $user = $c->user;
  my $params = $c->request->parameters;
  my ($content_type) = grep { $self->{map}->{$_} } @{$c->request->accepted_content_types};

  if( $content_type && $content_type !~ /json/ ) {
    return $self->status_bad_request( $c, message => "Incorrect content-type" );
  }

  $params->{user_id} = $user->id;
  my $ret = $c->model( 'Users' )->session_list( $params );

  my %resp = (
    status => 'success',
    time => sprintf( "%.3f", tv_interval( $t ) ),
    sessions => $ret
  );
$c->log->debug( Dumper(\%resp) );
  return $self->status_ok( $c, entity => \%resp );
}

sub _handle_create {
  my ( $self, $c, $params ) = @_;

  my $t = [gettimeofday];
  my $user = $c->user;

  my ($content_type) = grep { $self->{map}->{$_} } @{$c->request->accepted_content_types};

  if( $content_type && $content_type !~ /json/ ) {
    return $self->status_bad_request( $c, message => "Incorrect content-type" );
  }

  $params->{user_id} = $user->id;
  my $ret = $c->model( 'Users' )->session_new( $params );

$c->log->debug( Dumper($ret) );

  my %resp = (
    status => 'success',
    time => sprintf( "%.3f", tv_interval( $t ) ),
    skey => $ret
  );
  return $self->status_ok( $c, entity => \%resp );
}

sub create : Local : ActionClass('REST') { }
sub create_GET  :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_handle_create( $c, $c->request->parameters );
}
sub create_POST :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_handle_create( $c, $self->_post_params( $c ) );
}

sub update : Local : ActionClass('REST') { }
sub update_GET  :Args(0){
  my ( $self, $c ) = @_;

  my $user = $c->user;
  my $params = $c->request->parameters;
  my ($content_type) = grep { $self->{map}->{$_} } @{$c->request->accepted_content_types};

  if( $content_type && $content_type !~ /json/ ) {
    return $self->status_bad_request( $c, message => "Incorrect content-type" );
  }

  $params->{user_id} = $user->id;
  my $ret = $c->model( 'Users' )->session_update( $params );

  $c->log->debug( Dumper($ret) );

  return $self->status_ok( $c, entity => { status => $ret } );
}

sub fetch_meta : Local : ActionClass('REST') { }
sub fetch_meta_GET  :Args(0){
  my ( $self, $c ) = @_;

  my $t = [gettimeofday];

  my $user = $c->user;
  my $params = $c->request->parameters;
  my ($content_type) = grep { $self->{map}->{$_} } @{$c->request->accepted_content_types};

  if( $content_type && $content_type !~ /json/ ) {
    return $self->status_bad_request( $c, message => "Incorrect content-type" );
  }

  $params->{user_id} = $user->id;
  my $ret = $c->model( 'Users' )->session_fetch_meta( $params );

  $c->log->debug( Dumper($ret) );

  my %resp = (
    status => 'success',
    time => sprintf( "%.3f", tv_interval( $t ) ),
    session => $ret
  );

  return $self->status_ok( $c, entity => \%resp );
}

sub fetch : Local : ActionClass('REST') { }
sub fetch_GET  :Args(0){
  my ( $self, $c ) = @_;

  my $t = [gettimeofday];

  my $user = $c->user;
  my $params = $c->request->parameters;
  my ($content_type) = grep { $self->{map}->{$_} } @{$c->request->accepted_content_types};

  if( $content_type && $content_type !~ /json/ ) {
    return $self->status_bad_request( $c, message => "Incorrect content-type" );
  }

  $params->{user_id} = $user->id;
  my $ret = $c->model( 'Users' )->session_fetch( $params );

  $c->log->debug( Dumper($ret) );

  my %resp = (
    status => 'success',
    time => sprintf( "%.3f", tv_interval( $t ) ),
    session => $ret
  );

  return $self->status_ok( $c, entity => \%resp );
}

sub delete : Local : ActionClass('REST') { }
sub delete_GET  :Args(0){
  my ( $self, $c ) = @_;

  my $user = $c->user;
  my $params = $c->request->parameters;
  my ($content_type) = grep { $self->{map}->{$_} } @{$c->request->accepted_content_types};

  if( $content_type && $content_type !~ /json/ ) {
    return $self->status_bad_request( $c, message => "Incorrect content-type" );
  }

  $params->{user_id} = $user->id;
  my $ret = $c->model( 'Users' )->session_delete( $params );

  $c->log->debug( Dumper($ret) );

  return $self->status_ok( $c, entity => { status => $ret } );
}

sub serialize : ActionClass('Serialize') {}
sub end :Private {
  my ($self, $c) = @_;
  $c->forward('serialize');
}


=head1 AUTHOR

David Santamauro

=head1 LICENSE

This library is free software. You can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

__PACKAGE__->config(
  'default'   => 'application/json',
  'stash_key' => 'rest',
  'map'       => {

         'text/xml'           => 'XML::Passthrough',
         'application/json'   => 'JSON',
         'text/x-json'        => 'JSON',

         'text/html'          => 'YAML::HTML',
         'text/x-data-dumper' => [ 'Data::Serializer', 'Data::Dumper' ],
         'text/x-data-denter' => [ 'Data::Serializer', 'Data::Denter' ],
         'text/x-data-taxi'   => [ 'Data::Serializer', 'Data::Taxi'   ],
         'application/x-storable'   => [ 'Data::Serializer', 'Storable' ],
         'application/x-freezethaw' => [ 'Data::Serializer', 'FreezeThaw' ],
         'text/x-config-general'    => [ 'Data::Serializer', 'Config::General' ],
         'text/x-php-serialization' => [ 'Data::Serializer', 'PHP::Serialization' ],
        }
);


__PACKAGE__->meta->make_immutable;

1;
