package Alexandria::CDWS::Controller::Text;
use Moose;
use namespace::autoclean;

use Data::Dumper;
use XML::XML2JSON;

BEGIN {extends 'Catalyst::Controller::REST'; }

=head1 NAME

Alexandria::CDWS::Controller::Text - base text controller


=head1 DESCRIPTION

Search Catalyst Controller.

=head1 METHODS

=cut

use Data::Dumper;

sub _post_params {
  my ( $self, $c ) = @_;
  my %params;

  if($c->req->content_type eq 'application/x-www-form-urlencoded') {
    my $s = $c->request->parameters;
    foreach my $key ( keys %{$s} ) {
      my $val = $s->{$key};
      $val =~ s/^\s+//;
      $val =~ s/\s+$//;
      $params{$key} = $val;
    }
  }
  else {
    my $data;
    if(ref($c->request->body) =~ /File/) {
      my $tmp = $c->request->body;
      $tmp->seek(0,0);
      $data = do { local $/; <$tmp> };
    }
    elsif($c->request->body) {
      $data = $c->request->body;
    }
    else {
      my @a = ();
      my $h = $c->request->parameters;
      foreach my $key ( keys %{$h} ) {
        if(ref($h->{$key}) =~ /ARRAY/) {
          foreach my $f ( @{$h->{$key}} ) {
            push(@a, "$key=$f");
          }
        }
        else { push(@a, "$key=$h->{$key}"); }
      }
      $data = join('&', @a);
    }
    my $str = uri_unescape($data);
    $str =~ s/^\s+//;
    $str =~ s/\s+$//;
    my @a = split(/\&/, $str);
    foreach my $e (@a) {
      my ($key, $val) = split(/=/, $e);
      $params{$key} = $val;
    }
  }
  return \%params;
}

sub _process_request {
  my ( $self, $c, $params ) = @_;

  my $data = $params->{ucid};
  my $cd = $params->{containers};
  my $inventory = $params->{inventory};

  my @containers = ();
  if( $cd ) {
    $cd =~ s/^\s+//;
    $cd =~ s/\s+$//;
    @containers = split( ',', $cd );
  }

  my $user = $c->user;
  my ($content_type) = grep { $self->{map}->{$_} } @{$c->request->accepted_content_types};

  if(!$data) {
    return $self->status_bad_request($c, message => 'insufficient parameters');
  }

  $data =~ s/\n/ /g;  # remove all new lines
  $data =~ s/,/ /g;   # remove all commas
  $data =~ s/\s+/ /g; # tighten up spaces
  $data =~ s/^\s+//;
  $data =~ s/\s+$//;
  my @ucids = split(/\s+/, $data);

  $c->response->status(200);

  if( $inventory ) {
    my $xml = $c->model( 'Inventory' )->get_inventory_xml( $ucids[0] );
# $c->log->debug( $xml );
    $c->response->content_type('text/xml');
    $c->response->body( $xml );
    return 1;
  }

  my $LENGTH_CHECK_STRING =
    "<?xml version='1.0' encoding='utf-8' ?>
     <patent-document></patent-document>";

  # always text/xml ... is it? no, customers assume JSON is also supported
  if( ($content_type && $content_type =~ /json/)  || ($params->{'content-type'} && $params->{'content-type'} =~ /json/) ) {
    $c->response->content_type('application/json');
    $c->model( $user->rdb_model  )->fetch_xml( $c, \@ucids,
      sub {
        my ($c, $xml) = @_;
        my $obj = XML::XML2JSON->new();
        if( !$xml || length( $xml ) < length( $LENGTH_CHECK_STRING ) ) {
          # we'll at least get a timestamp and notion of frequency
          $c->log->error( sprintf("EMPTY XML: %s", ( $xml ) ? $xml : 'undef' ) );
          my $s = $LENGTH_CHECK_STRING;
          $s =~ s/^<\?.*\?>//;
          $c->response->write( $obj->convert( $s ) );
          return;
        }
        $c->response->write( $obj->convert( $xml ) );
      },
      \@containers
    );
    return 1;
  }

  $c->response->content_type('text/xml');
  $c->response->write("<?xml version='1.0' encoding='utf-8' ?>\n");
  $c->response->write('<patent-documents>') if (scalar(@ucids) > 1);
  $c->model( $user->rdb_model  )->fetch_xml( $c, \@ucids,
    sub {
      my ($c, $xml) = @_;
      if( !$xml || length( $xml ) < length( $LENGTH_CHECK_STRING ) ) {
        # we'll at least get a timestamp and notion of frequency
        $c->log->error( sprintf("EMPTY XML: %s", ( $xml ) ? $xml : 'undef' ) );
        my $s = $LENGTH_CHECK_STRING;
        $s =~ s/^<\?.*\?>//;
        $c->response->write( $s );
        return;
      }
      $c->response->write( $xml );
    },
    \@containers
  );
  $c->response->write('</patent-documents>') if (scalar(@ucids) > 1);
  return 1;
}

sub fetch : Local : ActionClass('REST') { }
sub fetch_GET  :Args(0){
  my ( $self, $c ) = @_;
  return $self->_process_request ($c, $c->request->parameters );
}

sub fetch_POST :Args(0) {
  my ( $self, $c ) = @_;
  return $self->_process_request( $c, $self->_post_params( $c ) );
}


=head2 end

=cut

sub serialize : ActionClass('Serialize') {}
sub end :Private {
  my ($self, $c) = @_;
  # No serialization ...
  # $c->forward('serialize');
}


=head1 AUTHOR

David Santamauro

=head1 LICENSE

This library is free software. You can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

__PACKAGE__->config(
  'default'   => 'text/xml',
  'stash_key' => 'rest',
  'map'       => {

         'text/xml'           => 'XML::Passthrough',
         'application/json'   => 'JSON::Passthrough',
         'text/x-json'        => 'JSON::Passthrough',

         'text/html'          => 'YAML::HTML',
         'text/x-data-dumper' => [ 'Data::Serializer', 'Data::Dumper' ],
         'text/x-data-denter' => [ 'Data::Serializer', 'Data::Denter' ],
         'text/x-data-taxi'   => [ 'Data::Serializer', 'Data::Taxi'   ],
         'application/x-storable'   => [ 'Data::Serializer', 'Storable' ],
         'application/x-freezethaw' => [ 'Data::Serializer', 'FreezeThaw' ],
         'text/x-config-general'    => [ 'Data::Serializer', 'Config::General' ],
         'text/x-php-serialization' => [ 'Data::Serializer', 'PHP::Serialization' ],
        }
);


__PACKAGE__->meta->make_immutable;

1;
