package Alexandria::CDWS::Controller::Update;
use Moose;
use namespace::autoclean;

BEGIN {extends 'Catalyst::Controller::REST'; }


=head1 NAME

Alexandria::Controller::Update - Catalyst Controller for Remote Client Updates

=head1 DESCRIPTION

Catalyst Controller.

=head1 METHODS

=cut

=head2 index

=cut
use Data::Dumper;
use Time::HiRes qw( gettimeofday tv_interval );

sub info : Local : ActionClass('REST') { }
sub info_GET  :Args(0){
  my ( $self, $c ) = @_;
  my $user = $c->user;

  my $t = [gettimeofday];

  my $params = $c->request->parameters;

  if( !$params->{load_id} ) {
    return $self->status_bad_request( $c, message => "Missing required load_id parameter" );
  }
  my $info = $c->model( $user->rdb_model )->loadid_information( $params->{load_id} );

$c->log->debug(Dumper($info));

  if( !$info->{load}->{load_source} ) {
    return $self->status_no_content( $c, message => "No information available for $params->{load_id}" );
  }


  my %ret = (
    load_info => $info->{load},
    time => sprintf("%.3f", tv_interval( $t ) ),
    status => 'success'
  );
  return $self->status_ok( $c, entity => \%ret );
}

sub check : Local : ActionClass('REST') { }
sub check_GET  :Args(0){
  my ( $self, $c ) = @_;
  my $params = $c->request->parameters;
  my $user = $c->user;

  my $t = [gettimeofday];

  if(!defined($params->{max_client_modified_load_id})) {
    $c->log->warn(__PACKAGE__ . "::check(): ($user->{name}) - missing required max_client_modified_load_id parameter");
    return $self->status_bad_request( $c, message => 'missing required max_client_modified_load_id parameter' );
  }
  my $available_load_info = $c->model( $user->rdb_model )->loads_pending( $params->{max_client_modified_load_id}, $params->{limit} );

  my %ret = (
    load_info => $available_load_info,
    time => sprintf("%.3f", tv_interval( $t ) ),
    status => 'success'
  );
  return $self->status_ok( $c, entity => \%ret );
}

sub next : Local : ActionClass('REST') { }
sub next_GET  :Args(0){
  my ( $self, $c ) = @_;
  my $params = $c->request->parameters;
  my $user = $c->user;

  my $t = [gettimeofday];

  if(!defined($params->{max_client_modified_load_id})) {
    $c->log->warn(__PACKAGE__ . "::get_next_load(): ($user->{name}) - missing required max_client_modified_load_id parameter");
    return $self->status_bad_request( $c, message => 'missing required max_client_modified_load_id parameter' );
  }
  my $available_load_info = $c->model( $user->rdb_model )->loads_pending( $params->{max_client_modified_load_id} );

  my %ret = (
    load_info => shift( @{$available_load_info} ),
    time => sprintf("%.3f", tv_interval( $t ) ),
    status => 'success'
  );
  return $self->status_ok( $c, entity => \%ret );
}

=head2 end

=cut

sub serialize : ActionClass('Serialize') {}
sub end :Private {
  my ($self, $c) = @_;
  $c->forward('serialize');
}


=head1 AUTHOR

David Santamauro

=head1 LICENSE

This library is free software. You can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

__PACKAGE__->config(
  'default'   => 'application/json',
  'stash_key' => 'rest',
  'map'       => {
         'text/html'          => 'YAML::HTML',
         'text/xml'           => 'XML::Simple',
         'application/json'   => 'JSON::Alexandria',
         'text/x-json'        => 'JSON::Alexandria',
         'text/x-data-dumper' => [ 'Data::Serializer', 'Data::Dumper' ],
         'text/x-data-denter' => [ 'Data::Serializer', 'Data::Denter' ],
         'text/x-data-taxi'   => [ 'Data::Serializer', 'Data::Taxi'   ],
         'application/x-storable'   => [ 'Data::Serializer', 'Storable' ],
         'application/x-freezethaw' => [ 'Data::Serializer', 'FreezeThaw' ],
         'text/x-config-general'    => [ 'Data::Serializer', 'Config::General' ],
         'text/x-php-serialization' => [ 'Data::Serializer', 'PHP::Serialization' ],
        }
);


__PACKAGE__->meta->make_immutable;

1;
