package Alexandria::CDWS::Error;

use Moose;
  with 'MooseX::Emulate::Class::Accessor::Fast';
use Scalar::Util qw/blessed/;
use MRO::Compat;
use namespace::autoclean;

sub execute {
    my $c = shift;

    my $conf = $c->config->{error};

    return $c->next::method(@_) unless defined $conf->{disabled};

    local $SIG{__DIE__} = sub {
        my $error = shift;

        # ignore if the error is a Tree::Simple object
        # because FindByUID uses an internal die several times per request
        return if ( blessed($error) && $error->isa('Tree::Simple') );

        die $error;
    };
    return $c->next::method(@_);
}

sub finalize_error {
    my $c = shift;

    $c->next::method(@_);
    $c->res->{body} = "Oh my goodness, it's an error!";
}

sub setup {
    my $c = shift;
    $c->next::method(@_);
}

1;
__END__

=pod

=head1 NAME

Alexandria::CDWS::Error - Display a stack trace on the debug screen

=head1 SYNOPSIS

    use Catalyst qw/-Debug CDWSError/;

=head1 DESCRIPTION

This plugin will enhance the standard Catalyst debug screen by including
a stack trace of your appliation up to the point where the error occurred.
Each stack frame is displayed along with the package name, line number, file
name, and code context surrounding the line number.

This plugin is only active in -Debug mode by default, but can be enabled by
setting the C<enable> config option.

=head1 CONFIGURATION

Configuration is optional and is specified in MyApp->config->{stacktrace}.

=head2 enable

Allows you forcibly enable or disalbe this plugin, ignoring the current 
debug setting. If this option is defined, its value will be used.

=head2 context

The number of context lines of code to display on either side of the stack
frame line.  Defaults to 3.

=head2 reverse

By default, the stack frames are shown in from "top" to "bottom"
(newest to oldest). Enabling this option reverses the stack frames so they will
be displayed "bottom" to "top", or from the callers perspective.

=head2 verbose

This option sets the amount of stack frames you want to see in the stack
trace.  It defaults to 0, meaning only frames from your application's
namespace are shown.  You can use levels 1 and 2 for deeper debugging.

If set to 1, the stack trace will include frames from packages outside of
your application's namespace, but not from most of the Catalyst internals.
Packages ignored at this level include:

    Catalyst
    Catalyst::Action
    Catalyst::Base
    Catalyst::Dispatcher
    Catalyst::Engine::*
    Alexandria::CDWS::Error
    Catalyst::Plugin::Static::Simple
    NEXT
    main

If set to 2, the stack trace will include frames from everything except this
module.

=head1 INTERNAL METHODS

The following methods are extended by this plugin.

=over 4

=item execute

In execute, we create a local die handler to generate the stack trace.

=item finalize_error

In finalize_error, we inject the stack trace HTML into the debug screen below
the error message.

=item setup

=back

=head1 SEE ALSO

L<Catalyst>

=head1 AUTHORS

Andy Grundman, <andy@hybridized.org>

Matt S. Trout, <mst@shadowcatsystems.co.uk>

David Santamauro, <david.santamauro@fairviewresearch.com>

=head1 THANKS

The authors of L<CGI::Application::Plugin::DebugScreen>, from which a lot of
code was used.

=head1 COPYRIGHT

Copyright (c) 2005 - 2009
the Alexandria::CDWS::Error L</AUTHORS>
as listed above.

=head1 LICENSE

This program is free software, you can redistribute it and/or modify it
under the same terms as Perl itself.

=cut
