package Alexandria::CDWS::Model::Attachments;

use strict;
use base 'Catalyst::Model::DBI';
use Data::Dumper;

sub fetch_path {
  my $self = shift;
  my $path = shift;
  my $abs = sprintf("%s/%s", $self->{root}, $path);
}

sub find_primary {
  my $self = shift;
  my $pkey = shift;

  my $dbh = $self->dbh;
  my $sql = <<EOS;
    select id
    from attachments.t_primaries
    where pkey = ?
EOS
  my $sth = $dbh->prepare($sql);
  $sth->execute($pkey);
  my ($id) = $sth->fetchrow_array();
  return $id;
}

sub find {
  my $self = shift;
  my $pkey = shift;
  my $file = shift;
  my $sql = <<EOS;
    select count(p.pkey)
    from attachments.t_attachments a
      inner join attachments.t_primaries p on (a.pref=p.id)
    where p.pkey = ?
      and a.filename=?
EOS

  my $dbh = $self->dbh;
  my $sth = $dbh->prepare($sql);
  $sth->execute($pkey, $file);
  my ($count) = $sth->fetchrow_array();
  return $count;
}

sub _store_attachment_db {
  my ($self, $params) = @_;
  my $dbh = $self->dbh;

  my $id = $self->find_primary($params->{pkey});
  if($id) {
    if($self->find($params->{pkey}, $params->{file})) {
      # be silent ...
      return 1;
    }
  }
  else {
    my $sth = $dbh->prepare("insert into attachments.t_primaries (pkey, path) values(?, ?)");
    $sth->execute($params->{pkey}, $params->{path});
    $id = $dbh->last_insert_id(undef,undef,undef,undef,{sequence=>'attachments.seq_prim'});
  }

  if(!$id) { die "failed to establish t_primaries :id for $params->{pkey}"; }

  # now store the actual image in t_attachments
  my $sth = $dbh->prepare("insert into attachments.t_attachments (pref, size, filename) values(?,?,?)");
  $sth->execute($id, $params->{size}, $params->{file});
  $sth->finish();
}

use File::Path;
sub _store_attachment_fs {
  my ($self, $params) = @_;

  my $path = $params->{fs};
  my $file = "$params->{fs}/$params->{file}";
  if(!(-d $path)) {
    eval { mkpath($path); };
    if($@) { die "$file: $!\n"; }
  }
  my $fh;
  open($fh, ">$file") || die "$file: $!\n";
  syswrite($fh, $params->{data}, $params->{size});
  close($fh);
}

sub _remove_attachment_fs {
  my ($self, $params) = @_;
  my $file = "$params->{fs}/$params->{file}";
  unlink($file); # silently
}

sub store {
  my $self = shift;
  my $params = shift;
#  my %params = (
#    pkey => pkey       # column: t_primaries.pkey
#    ucid => $ucid      # column: t_primaries.ucid
#    path => $location, # column: t_primaries.path
#    size => $size,     # column: t_attachments.size
#    file => $file,     # column: t_attachments.file
#    fs   => $fs (without file)
#    data => $data
#  );

  # UGLY ... web server storing data
  # TODO: do database checks first before
  # writing to the file system
  eval { $self->_store_attachment_fs($params); };
  if($@) { die $@; }

  eval { $self->_store_attachment_db($params); };
  if($@) {
    $self->_remove_attachment_fs($params);
    die $@;
  }
}

sub _build_public_entries {
  my ($self, $pref) = @_;
  my %h = (
    pkey => $pref->{pkey},
    filename => $pref->{filename},
    size  => $pref->{size},
    path  => $pref->{path} . '/' . $pref->{filename},
  );

  my %exttbl = (
    tif => 'image/tiff',
    tiff => 'image/tiff',
    xml => 'text/xml',
    mol => 'application/mol',
    nb  => 'application/nb',
    cdx => 'application/cdx',
    pdf => 'application/pdf',
    jpg => 'image/jpeg',
    png => 'image/png'
  );
  my ($ext) = ($pref->{filename} =~ /.*\.([A-Za-z]+)$/);
  $h{media} = ($exttbl{lc($ext)}) ? $exttbl{lc($ext)} : 'application/unknown';
  return \%h
}

sub has_attachments {
  my $self = shift;
  my $ucid = shift;

  my $sql = "select count(*)
             from attachments.t_attachments a
               inner join attachments.t_primaries p on (a.pref=p.id)
             where p.id in (select attachments.f_ucid2attachid(?))";

  my $dbh = $self->dbh;
  my $sth = $dbh->prepare($sql);
  $sth->execute($ucid);
  my ($count) = $sth->fetchrow_array();
  $sth->finish;
  return ($count) ? $count : 0;
}

sub list {
  my $self = shift;
  my $ucid = shift;
  my $c = $self->{c}; # read-only

  # with PrintError=1 (connection attributes), DBD::Pg prints
  # warnings as well as throws exceptions.
  local $SIG{'__WARN__'} = sub {
    warn __PACKAGE__ . ': (wow) ' . $_[0]
      if $_[0] !~ /ERROR:.*?not found/;
  };

  my @attachments = ();

  my $sql = <<EOS;
    select p.pkey,p.path,a.*
    from attachments.t_attachments a
      inner join attachments.t_primaries p on (a.pref=p.id)
    where p.id in (select attachments.f_ucid2attachid(?))
    order by a.filename
EOS

  eval {
    my $dbh = $self->dbh;
    my $sth = $dbh->prepare($sql);
    $sth->execute($ucid);
    while(my $ref = $sth->fetchrow_hashref()) {
      push(@attachments, $self->_build_public_entries($ref));
    }
  };
  if($@) {
    my $e = $@;
    if($e =~ /ERROR:.*?not found/) { } # no op
    else { $c->log->error( $e ); }
  }
  return \@attachments;
}

sub ping {
  my $self = shift;
  my $dbh = $self->dbh;
  my $ret = $dbh->pg_ping();
  if($ret < 1) { return 0; }
  return $ret;
}

sub ACCEPT_CONTEXT {
  my ($self, $c, @args) = @_;
  $self = bless( { %{$self}, c => $c }, ref($self) );
  return $self;
}

__PACKAGE__->config(
  AutoCommit => q{1},
);

1;

__END__

