package Alexandria::CDWS::Model::Solr;

use utf8;
use strict;
use warnings;

use Moose;
use Moose::Util::TypeConstraints;
use Alexandria::WebService::Solr;
use Alexandria::WebService::Solr::Response;
use namespace::autoclean;
use Data::Dumper;
extends 'Catalyst::Model';

has 'server' => (
  is      => 'ro',
  isa     => 'Str',
  default => 'http://solr.alexandria.com:8080/alexandria-index/alexandria',
);

has 'base' => (
  is      => 'ro',
  isa     => 'Str',
  default => 'http://solr.alexandria.com:8080/alexandria-index',
);

has 'options' => (
  is      => 'rw',
  isa     => 'HashRef',
  default => sub { {} },
);

has 'solr' => (
  is         => 'rw',
  isa        => 'Alexandria::WebService::Solr',
  handles    => qr{^[^_].*},
  lazy_build => 1
);

sub _build_solr {
  my $self = shift;
  return Alexandria::WebService::Solr->new( $self->server, $self->options );
}

# SCHEMA
sub schema {
  my ($self, $path, $params) = @_;

  my $url = $self->{base} . '/' . $self->{collection};

  if($path) { $url .= "/$path";  }
  else      { $url .= "/schema"; } 

  my @a = ();
  foreach my $key ( sort keys %{$params} ) { push(@a, "$key=$params->{$key}"); }
  $url .= sprintf("?%s", join("&", @a));

  my $agent = $self->solr->agent;
  # all schema requests are JSON
  return Alexandria::WebService::Solr::Response->new( $agent->get($url) );
}

# gather application information for the authorized core
sub core_statistics {
  my ($self, $c) = @_;
  my %urls = (
    details => $self->server . "/admin/luke?show=schema&show=index&wt=json",
    replication => $self->server . "/replication?command=details&wt=json",
    core => $self->base . "/admin/cores?core=" . $self->core . "&wt=json",
  );

  my %ret = ();
  foreach my $key ( keys %urls ) {
    my $url = $urls{$key};
    my $agent = $self->solr->agent;
    my $response = Alexandria::WebService::Solr::Response->new( $agent->get($url) );

    if($key eq 'core') {
      $ret{$key} = $response->content->{status}->{$self->core};
    }
    else {
      $ret{$key} = $response->content;
    }
  }
  return \%ret;
}


sub idx_stats {
  my ($self, $params, $rdb) = @_;

  my %ret = ();

  if($params && $params->{component} && $params->{component} eq 'queue') {
    return $rdb->_index_stats( $params );
  }
  elsif($params && $params->{component} && $params->{component} eq 'core') {
    return $self->core_statistics();
  }
  elsif($params && $params->{component} && $params->{component} eq 'cloud') {
    return \%ret;
  }
  else {
    $ret{queue} = $rdb->_index_stats( $params );
    $ret{core} = $self->core_statistics();
    return \%ret;
  }
}


__PACKAGE__->config(
  options => {
    autocommit => 1,
  }
);


=head1 NAME

Alexandria::Model::Solr - Catalyst Model

=head1 DESCRIPTION

Alexandria::Model::Solr is a simple sub-class of Catalyst::Model.

=head1 AUTHOR

David Santamauro E<lt>david.santamauro@fairviewresearch.comE<gt>

=head1 LICENSE

This library is free software. You can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

__PACKAGE__->meta->make_immutable;

1;

